#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect
import itk
import numpy as np


def create_spect_simu(sim, paths, number_of_threads=1):
    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.number_of_threads = number_of_threads
    ui.random_seed = 123456

    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    kBq = 1000 * Bq

    # world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]
    world.material = "G4_AIR"

    # spect head (debug mode = very small collimator)
    spect, crystal = gate_spect.add_ge_nm67_spect_head(
        sim, "spect", collimator_type=False, debug=False
    )
    psd = 6.11 * cm
    spect.translation = [0, 0, -(20 * cm + psd)]

    # waterbox
    waterbox = sim.add_volume("Box", "waterbox")
    waterbox.size = [15 * cm, 15 * cm, 15 * cm]
    waterbox.material = "G4_WATER"
    blue = [0, 1, 1, 1]
    waterbox.color = blue

    # physic list
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = False
    cuts = p.production_cuts
    cuts.world.gamma = 10 * mm
    cuts.world.electron = 10 * mm
    cuts.world.positron = 10 * mm
    cuts.world.proton = 10 * mm

    cuts.spect.gamma = 0.1 * mm
    cuts.spect.electron = 0.01 * mm
    cuts.spect.positron = 0.1 * mm

    # default source for tests
    activity = 30 * kBq
    beam1 = sim.add_source("GenericSource", "beam1")
    beam1.mother = waterbox.name
    beam1.particle = "gamma"
    beam1.energy.mono = 140.5 * keV
    beam1.position.type = "sphere"
    beam1.position.radius = 3 * cm
    beam1.position.translation = [0, 0, 0 * cm]
    beam1.direction.type = "momentum"
    beam1.direction.momentum = [0, 0, -1]
    # beam1.direction.type = 'iso'
    beam1.activity = activity / ui.number_of_threads

    beam2 = sim.add_source("GenericSource", "beam2")
    beam2.mother = waterbox.name
    beam2.particle = "gamma"
    beam2.energy.mono = 140.5 * keV
    beam2.position.type = "sphere"
    beam2.position.radius = 3 * cm
    beam2.position.translation = [18 * cm, 0, 0]
    beam2.direction.type = "momentum"
    beam2.direction.momentum = [0, 0, -1]
    # beam2.direction.type = 'iso'
    beam2.activity = activity / ui.number_of_threads

    beam3 = sim.add_source("GenericSource", "beam3")
    beam3.mother = waterbox.name
    beam3.particle = "gamma"
    beam3.energy.mono = 140.5 * keV
    beam3.position.type = "sphere"
    beam3.position.radius = 1 * cm
    beam3.position.translation = [0, 10 * cm, 0]
    beam3.direction.type = "momentum"
    beam3.direction.momentum = [0, 0, -1]
    # beam3.direction.type = 'iso'
    beam3.activity = activity / ui.number_of_threads

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    # get crystal volume by looking for the word crystal in the name
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    hc.mother = crystal.name
    print("Crystal :", crystal.name)
    hc.output = paths.output / "test028.root"
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "TrackVolumeCopyNo",
        "PostStepUniqueVolumeID",
        "PreStepUniqueVolumeID",
        "GlobalTime",
        "KineticEnergy",
        "ProcessDefinedStep",
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.mother = crystal.name
    sc.input_digi_collection = "Hits"
    sc.policy = "EnergyWinnerPosition"
    # sc.policy = 'EnergyWeightedCentroidPosition'
    sc.skip_attributes = ["KineticEnergy", "ProcessDefinedStep", "KineticEnergy"]
    sc.output = hc.output

    # EnergyWindows
    cc = sim.add_actor("DigitizerEnergyWindowsActor", "EnergyWindows")
    cc.mother = crystal.name
    cc.input_digi_collection = "Singles"
    cc.channels = [
        {"name": "scatter", "min": 114 * keV, "max": 126 * keV},
        {"name": "peak140", "min": 126 * keV, "max": 154.55 * keV},
        {
            "name": "spectrum",
            "min": 0 * keV,
            "max": 5000 * keV,
        },  # should be strictly equal to 'Singles'
    ]
    cc.output = hc.output

    """
        The order of the actors is important !
        1. Hits
        2. Singles
        3. EnergyWindows
    """

    # sec = gate.g4_units('second')
    # sim.run_timing_intervals = [[0, 0.5 * sec], [0.5 * sec, 1 * sec]]

    return spect


def test_add_proj(sim, paths):
    mm = gate.g4_units("mm")
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    # 2D binning projection
    proj = sim.add_actor("HitsProjectionActor", "Projection")
    proj.mother = crystal.name
    # we set two times the spectrum channel to compare with Gate output
    proj.input_digi_collections = ["spectrum", "scatter", "peak140", "spectrum"]
    proj.spacing = [4.41806 * mm, 4.41806 * mm]
    proj.size = [128, 128]
    # proj.plane = 'XY' # not implemented yet # FIXME
    proj.output = paths.output / "proj028.mhd"
    # by default, the origin of the images are centered
    # set to False here to keep compatible with previous version
    proj.origin_as_image_center = False
    return proj


def test_spect_hits(output, paths, version="2"):
    # stat
    gate.warning("Compare stats")
    stats = output.get_actor("Stats")
    print(stats)
    print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
    stats.counts.run_count = 1  # force to 1
    stats_ref = gate.read_stat_file(paths.gate_output / f"stat{version}.txt")
    is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.07)

    # Compare root files
    print()
    gate.warning("Compare hits")
    gate_file = paths.gate_output / f"hits{version}.root"
    hc_file = output.get_actor("Hits").user_info.output
    print(hc_file)
    checked_keys = [
        {"k1": "posX", "k2": "PostPosition_X", "tol": 1.7, "scaling": 1},
        {"k1": "posY", "k2": "PostPosition_Y", "tol": 1.3, "scaling": 1},
        {"k1": "posZ", "k2": "PostPosition_Z", "tol": 0.9, "scaling": 1},
        {"k1": "edep", "k2": "TotalEnergyDeposit", "tol": 0.001, "scaling": 1},
        {"k1": "time", "k2": "GlobalTime", "tol": 0.01, "scaling": 1e-9},
    ]
    is_ok = (
        gate.compare_root2(
            gate_file,
            hc_file,
            "Hits",
            "Hits",
            checked_keys,
            paths.output / f"test028_{version}_hits.png",
            n_tol=4,
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.warning("Compare singles")
    gate_file = paths.gate_output / f"hits{version}.root"
    hc_file = output.get_actor("Singles").user_info.output
    checked_keys = [
        {"k1": "globalPosX", "k2": "PostPosition_X", "tol": 1.8, "scaling": 1},
        {"k1": "globalPosY", "k2": "PostPosition_Y", "tol": 1.3, "scaling": 1},
        {"k1": "globalPosZ", "k2": "PostPosition_Z", "tol": 0.2, "scaling": 1},
        {"k1": "energy", "k2": "TotalEnergyDeposit", "tol": 0.001, "scaling": 1},
    ]
    is_ok = (
        gate.compare_root2(
            gate_file,
            hc_file,
            "Singles",
            "Singles",
            checked_keys,
            paths.output / f"test028_{version}_singles.png",
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.warning("Compare singles and spectrum (must be strictly equal)")
    ref_file = output.get_actor("Singles").user_info.output
    hc_file = output.get_actor("EnergyWindows").user_info.output
    checked_keys = [
        {"k1": "PostPosition_X", "k2": "PostPosition_X", "tol": 0.001, "scaling": 1},
        {"k1": "PostPosition_Y", "k2": "PostPosition_Y", "tol": 0.001, "scaling": 1},
        {"k1": "PostPosition_Z", "k2": "PostPosition_Z", "tol": 0.001, "scaling": 1},
        {
            "k1": "TotalEnergyDeposit",
            "k2": "TotalEnergyDeposit",
            "tol": 0.001,
            "scaling": 1,
        },
    ]
    is_ok = (
        gate.compare_root2(
            ref_file,
            hc_file,
            "Singles",
            "spectrum",
            checked_keys,
            paths.output / f"test028_{version}_spectrum.png",
            n_tol=0.01,
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.warning("Compare scatter")
    hc_file = output.get_actor("EnergyWindows").user_info.output
    checked_keys = [
        {"k1": "globalPosX", "k2": "PostPosition_X", "tol": 20, "scaling": 1},
        {"k1": "globalPosY", "k2": "PostPosition_Y", "tol": 15, "scaling": 1},
        {"k1": "globalPosZ", "k2": "PostPosition_Z", "tol": 1.8, "scaling": 1},
        {"k1": "energy", "k2": "TotalEnergyDeposit", "tol": 0.2, "scaling": 1},
    ]
    is_ok = (
        gate.compare_root2(
            gate_file,
            hc_file,
            "scatter",
            "scatter",
            checked_keys,
            paths.output / f"test028_{version}_scatter.png",
            n_tol=15,
        )
        and is_ok
    )

    # Compare root files
    print()
    gate.warning("Compare peak")
    hc_file = output.get_actor("EnergyWindows").user_info.output
    checked_keys = [
        {"k1": "globalPosX", "k2": "PostPosition_X", "tol": 1.7, "scaling": 1},
        {"k1": "globalPosY", "k2": "PostPosition_Y", "tol": 1, "scaling": 1},
        {"k1": "globalPosZ", "k2": "PostPosition_Z", "tol": 0.21, "scaling": 1},
        {"k1": "energy", "k2": "TotalEnergyDeposit", "tol": 0.1, "scaling": 1},
    ]
    is_ok = (
        gate.compare_root2(
            gate_file,
            hc_file,
            "peak140",
            "peak140",
            checked_keys,
            paths.output / f"test028_{version}_peak.png",
            n_tol=2.1,
        )
        and is_ok
    )

    return is_ok


def test_spect_proj(output, paths, proj, version="3"):
    print()
    stats = output.get_actor("Stats")
    stats.counts.run_count = 1  # force to 1 to compare with gate result
    print(stats)
    stats_ref = gate.read_stat_file(paths.gate_output / f"stat{version}.txt")
    is_ok = gate.assert_stats(stats, stats_ref, 0.025)

    # compare images with Gate
    print()
    print("Compare images (old spacing/origin)")
    # read image and force change the offset to be similar to old Gate
    img = itk.imread(str(paths.output / "proj028.mhd"))
    spacing = np.array(proj.user_info.spacing)
    origin = spacing / 2.0
    origin[2] = 0.5
    spacing[2] = 1
    img.SetSpacing(spacing)
    img.SetOrigin(origin)
    itk.imwrite(img, str(paths.output / "proj028_offset.mhd"))
    is_ok = (
        gate.assert_images(
            paths.gate_output / f"projection{version}.mhd",
            paths.output / "proj028_offset.mhd",
            stats,
            tolerance=16,
            ignore_value=0,
            axis="y",
            sum_tolerance=1.6,
            fig_name=paths.output / f"proj028_{version}_offset.png",
        )
        and is_ok
    )

    # compare images with Gate
    if version == "3_blur":
        return is_ok
    print()
    print("Compare images (new spacing/origin")
    # read image and force change the offset to be similar to old Gate
    is_ok = (
        gate.assert_images(
            paths.output_ref / "proj028_ref.mhd",
            paths.output / "proj028.mhd",
            stats,
            tolerance=14,
            ignore_value=0,
            axis="y",
            sum_tolerance=1.5,
            fig_name=paths.output / f"proj028_{version}_no_offset.png",
        )
        and is_ok
    )

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import test019_linac_phsp_helpers as t
import opengate as gate

sim = t.init_test019(1)

sim.user_info.visu = True
sim.user_info.visu_type = "vrml"
sim.user_info.visu_filename = "geant4VisuFile.wrl"

source = sim.get_source_user_info("Default")
Bq = gate.g4_units("Bq")
source.activity = 1 * Bq

s = sim.dump_tree_of_volumes()
print(s)

t.run_test019(sim)



from .GenericSource import *
import opengate_core as g4
import math


class PencilBeamSource(GenericSource):
    """
    Pencil Beam source
    """

    type_name = "PencilBeamSource"

    @staticmethod
    def set_default_user_info(user_info):
        GenericSource.set_default_user_info(user_info)
        user_info.position.type = "disc"
        # additional parameters: direction
        # sigma, theta, epsilon, conv (0: divergent, 1: convergent)
        user_info.direction.partPhSp_x = [0, 0, 0, 0]
        user_info.direction.partPhSp_y = [0, 0, 0, 0]

    def __del__(self):
        pass

    def create_g4_source(self):
        return g4.GatePencilBeamSource()

    def __init__(self, user_info):
        super().__init__(user_info)
        self.__check_phSpace_params(self.user_info.direction.partPhSp_x)
        self.__check_phSpace_params(self.user_info.direction.partPhSp_y)

    def __check_phSpace_params(self, paramV):
        sigma = paramV[0]
        theta = paramV[1]
        epsilon = paramV[2]
        conv = paramV[3]
        pi = math.pi
        if epsilon == 0:
            raise ValueError(
                "Ellipse area is 0 !!! Check epsilon parameter in PencilBeamSource."
            )
        if pi * sigma * theta < epsilon:
            raise ValueError(
                f"pi*sigma*theta < epsilon. Provided values: sigma = {sigma}, theta = {theta}, epsilon = {epsilon}."
            )
        if conv not in [0, 1]:
            raise ValueError("convergence parameter can be only 0 or 1.")



from .ExceptionHandler import *
import os


class SimulationOutput:
    """
    FIXME
    """

    def __init__(self):
        self.simulation = None
        self.actors = {}
        self.sources = {}
        self.sources_by_thread = {}
        self.pid = os.getpid()
        self.ppid = os.getppid()
        self.current_random_seed = None

    def __del__(self):
        pass

    def store_actors(self, simulation_engine):
        self.actors = simulation_engine.actor_engine.actors

    def store_sources(self, simulation_engine):
        self.sources = {}
        s = {}
        source_engine = simulation_engine.source_engine
        ui = simulation_engine.simulation.user_info
        if ui.number_of_threads > 1 or ui.force_multithread_mode:
            th = {}
            self.sources_by_thread = [{}] * (ui.number_of_threads + 1)
            for source in source_engine.sources:
                n = source.user_info.name
                if n in th:
                    th[n] += 1
                else:
                    th[n] = 0
                self.sources_by_thread[th[n]][n] = source
        else:
            for source in source_engine.sources:
                s[source.user_info.name] = source
            self.sources = s

    def get_actor(self, name):
        if name not in self.actors:
            s = self.actors.keys
            gate.fatal(
                f'The actor "{name}" does not exist. Here is the list of actors: {s}'
            )
        return self.actors[name]

    def get_source(self, name):
        ui = self.simulation.user_info
        if ui.number_of_threads > 1 or ui.force_multithread_mode:
            return self.get_source_MT(name, 0)
        if name not in self.sources:
            s = self.sources.keys
            gate.fatal(
                f'The source "{name}" does not exist. Here is the list of sources: {s}'
            )
        return self.sources[name]

    def get_source_MT(self, name, thread):
        ui = self.simulation.user_info
        if ui.number_of_threads <= 1 and not ui.force_multithread_mode:
            gate.fatal(f"Cannot use get_source_MT in monothread mode")
        if thread >= len(self.sources_by_thread):
            gate.fatal(
                f"Cannot get source {name} with thread {thread}, while "
                f"there are only {len(self.sources_by_thread)} threads"
            )
        if name not in self.sources_by_thread[thread]:
            s = self.sources_by_thread[thread].keys
            gate.fatal(
                f'The source "{name}" does not exist. Here is the list of sources: {s}'
            )
        return self.sources_by_thread[thread][name]



import opengate_core as g4


class UIsessionSilent(g4.G4UIsession):
    """
    TODO
    """

    def __del__(self):
        pass

    def ReceiveG4cout(self, coutString):
        # print('HERE ', coutString)
        return 0

    def ReceiveG4cerr(self, cerrString):
        # print('HERE ', cerrString)
        return 0



import opengate as gate
from .ParticleFilter import ParticleFilter
from .KineticEnergyFilter import KineticEnergyFilter
from .TrackCreatorProcessFilter import TrackCreatorProcessFilter

filter_type_names = {ParticleFilter, KineticEnergyFilter, TrackCreatorProcessFilter}
filter_builders = gate.make_builders(filter_type_names)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.linac_elekta_synergy as gate_linac
import gatetools.phsp as phsp
import matplotlib.pyplot as plt

paths = gate.get_default_test_paths(__file__, "gate_test019_linac_phsp")


def init_test019(nt):
    # global log level
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.check_volumes_overlap = False
    ui.number_of_threads = nt
    ui.random_seed = 123456789
    print(ui)

    # units
    m = gate.g4_units("m")
    mm = gate.g4_units("mm")
    cm = gate.g4_units("cm")
    nm = gate.g4_units("nm")

    #  adapt world size
    world = sim.world
    world.size = [1 * m, 1 * m, 1 * m]

    # add a waterbox
    # waterbox = sim.add_volume('Box', 'Waterbox')
    # waterbox.size = [30 * cm, 30 * cm, 30 * cm]
    # waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
    # waterbox.material = 'G4_WATER'
    # waterbox.color = [0, 0, 1, 1]  # blue

    # add a linac
    linac = gate_linac.add_linac(sim, "linac")
    linac.translation = [0, 0, 0 * m]

    # virtual plane for phase space
    plane = sim.add_volume("Tubs", "phase_space_plane")
    plane.mother = world.name
    plane.material = "G4_AIR"
    plane.rmin = 0
    plane.rmax = 40 * mm
    plane.dz = 9 * cm  # half height
    plane.translation = [0, 0, -300 * mm - plane.dz]
    plane.color = [1, 0, 0, 1]  # red

    # e- source
    source = sim.add_source("GenericSource", "Default")
    Bq = gate.g4_units("Bq")
    MeV = gate.g4_units("MeV")
    source.particle = "e-"
    source.mother = f"{linac.name}_target"
    source.energy.type = "gauss"
    source.energy.mono = 6.7 * MeV
    source.energy.sigma_gauss = 0.077 * MeV
    source.position.type = "disc"
    source.position.radius = 2 * mm  # FIXME not really similar to GATE need sigma etc
    source.position.translation = [0, 0, 0.6 * mm]
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, -1]
    source.activity = 5000 * Bq / ui.number_of_threads

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # PhaseSpace Actor
    ta2 = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
    ta2.mother = plane.name
    ta2.attributes = [
        "KineticEnergy",
        "Weight",
        "PostPosition",
        "PrePosition",
        "ParticleName",
        "PreDirection",
        "PostDirection",
        "TimeFromBeginOfEvent",
        "GlobalTime",
        "LocalTime",
        "EventPosition",
    ]
    ta2.output = paths.output / "test019_hits.root"
    ta2.debug = False
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "gamma"
    ta2.filters.append(f)

    # phys
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = False
    cuts = p.production_cuts
    cuts.world.gamma = 1 * mm
    cuts.world.electron = 1 * mm
    cuts.world.positron = 1 * mm

    return sim


def run_test019(sim):
    # splitting
    linac = sim.get_volume_user_info("linac")
    s = f"/process/em/setSecBiasing eBrem {linac.name}_target 100 100 MeV"
    print(s)
    sim.apply_g4_command(s)

    # start simulation
    output = sim.start()

    # print results
    stats = output.get_actor("Stats")
    print(stats)

    h = output.get_actor("PhaseSpace")
    print(h)

    """
    not done yet:
    - missing several branch names in PhaseSpaceActor
    - no local/global for position
    - no policy options (all track single etc)
    """

    # check stats
    print()
    stats_ref = gate.read_stat_file(paths.gate_output / "output-writePhS-stat.txt")
    print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
    stats.counts.run_count = 1
    is_ok = gate.assert_stats(stats, stats_ref, 0.2)

    # compare the phsp tree
    print()
    fn1 = paths.gate_output / "output-PhS-g.root"
    fn2 = paths.output / "test019_hits.root"
    print("Reference gate tree : ", fn1)
    print("Checked Tree : ", fn2)
    data_ref, keys_ref, m_ref = phsp.load(fn1)
    data, keys, m = phsp.load(fn2)
    # find the good key's names
    keys1, keys2, scalings, tols = gate.get_keys_correspondence(keys_ref)
    # Do not check some keys
    tols[keys1.index("Weight")] = 0.002
    tols[keys1.index("Z")] = 0.09
    tols[keys1.index("Ekine")] = 0.1
    tols[keys1.index("Y")] = 1.6
    tols[keys1.index("X")] = 1.5
    tols[keys1.index("Z")] = 1.2
    # the Z position is not the same (plane is translated), and is fixed
    mm = gate.g4_units("mm")
    data[:, keys.index("PostPosition_Z")] += 297 * mm
    # perform the test
    is_ok = (
        gate.compare_trees(
            data_ref, keys_ref, data, keys, keys1, keys2, tols, scalings, scalings, True
        )
        and is_ok
    )

    # figure
    plt.suptitle(f"Values: {len(data_ref)} vs {len(data)}")
    # plt.show()
    fn = paths.output / "test019.png"
    plt.savefig(fn)
    print(f"Figure in {fn}")

    # this is the end, my friend
    gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4


class SourceManager:
    """
    Manage all the sources in the simulation.
    The function prepare_generate_primaries will be called during
    the main run loop to set the current time and source.
    """

    def __init__(self, simulation):
        # Keep a pointer to the current simulation
        self.simulation = simulation
        # List of run times intervals
        self.run_timing_intervals = None
        self.current_run_interval = None
        # List of sources user info
        self.user_info_sources = {}

    def __str__(self):
        """
        str only dump the user info on a single line
        """
        v = [v.name for v in self.user_info_sources.values()]
        s = f'{" ".join(v)} ({len(self.user_info_sources)})'
        return s

    def __del__(self):
        # print("del SourceManager")
        pass

    def dump(self):
        n = len(self.user_info_sources)
        s = f"Number of sources: {n}"
        for source in self.user_info_sources.values():
            a = f"\n {source}"
            s += gate.indent(2, a)
        return s

    def get_source_info(self, name):
        if name not in self.user_info_sources:
            gate.fatal(
                f"The source {name} is not in the current "
                f"list of sources: {self.user_info_sources}"
            )
        return self.user_info_sources[name]

    """def get_source(self, name):
        n = len(self.g4_thread_source_managers)
        if n > 0:
            gate.warning(f"Cannot get source in multithread mode, use get_source_MT")
            return None
        for source in self.sources:
            if source.user_info.name == name:
                return source.g4_source
        gate.fatal(
            f'The source "{name}" is not in the current '
            f"list of sources: {self.user_info_sources}"
        )

    def get_source_MT(self, name, thread):
        n = len(self.g4_thread_source_managers)
        if n == 0:
            gate.warning(f"Cannot get source in mono-thread mode, use get_source")
            return None
        i = 0
        for source in self.sources:
            if source.user_info.name == name:
                if i == thread:
                    return source.g4_source
                i += 1
        gate.fatal(
            f'The source "{name}" is not in the current '
            f"list of sources: {self.user_info_sources}"
        )"""

    def add_source(self, source_type, name):
        # check that another element with the same name does not already exist
        gate.assert_unique_element_name(self.user_info_sources, name)
        # init the user info
        s = gate.UserInfo("Source", source_type, name)
        # append to the list
        self.user_info_sources[name] = s
        # return the info
        return s



import opengate as gate
import opengate_core as g4
from box import Box


class SourceEngine(gate.EngineBase):
    """
    FIXME
    """

    # G4RunManager::BeamOn takes an int as input. The max cpp int value is currently 2147483647
    # Python manage int differently (no limit), so we need to set the max value here.
    max_int = 2147483647

    def __init__(self, source_manager):
        gate.EngineBase.__init__(self)

        # Keep a pointer to the current simulation
        self.source_manager = source_manager

        # List of run times intervals
        self.run_timing_intervals = None
        self.current_run_interval = None

        # List of sources (GateSource), for all threads
        self.sources = []

        # The source manager will be constructed at build (during ActionManager)
        # Its task is to call GeneratePrimaries and loop over the sources
        # For MT, the master_source_manager is the MasterThread
        # The g4_thread_source_managers list all master source for all threads
        self.g4_master_source_manager = None
        self.g4_thread_source_managers = []

        # internal variables
        self.g4_particle_table = None

        # Options dict for cpp SourceManager
        # will be set in create_g4_source_manager
        self.source_manager_options = Box()

    def __del__(self):
        if self.verbose_destructor:
            print("del SourceEngine")
        pass

    def initialize(self, run_timing_intervals):
        self.run_timing_intervals = run_timing_intervals
        gate.assert_run_timing(self.run_timing_intervals)
        uis = self.source_manager.user_info_sources
        if len(uis) == 0:
            gate.warning(f"No source: no particle will be generated")

    def initialize_actors(self, actors):
        actors = [actors[a] for a in actors]
        self.g4_master_source_manager.SetActors(actors)

    def create_master_source_manager(self):
        # create particles table # FIXME in physics ??
        self.g4_particle_table = g4.G4ParticleTable.GetParticleTable()
        self.g4_particle_table.CreateAllParticles()
        # create the master source for the masterThread
        self.g4_master_source_manager = self.create_g4_source_manager(False)
        return self.g4_master_source_manager

    def create_g4_source_manager(self, append=True):
        """
        This is called by all threads
        This object is needed here, because it can only be
        created after physics initialization
        """
        ms = g4.GateSourceManager()
        # create all sources for this source manager (for all threads)
        for vu in self.source_manager.user_info_sources.values():
            source = gate.new_element(vu, self.source_manager.simulation)
            ms.AddSource(source.g4_source)
            source.initialize(self.run_timing_intervals)
            self.sources.append(source)
        # taking __dict__ allow to consider the class SimulationUserInfo as a dict
        sui = self.source_manager.simulation.user_info.__dict__
        # warning: copy the simple elements from this dict (containing visu or verbose)
        for s in sui:
            if "visu" in s or "verbose_" in s:
                self.source_manager_options[s] = sui[s]
        ms.Initialize(self.run_timing_intervals, self.source_manager_options)
        # keep pointer to avoid deletion
        if append:
            self.g4_thread_source_managers.append(ms)
        return ms

    def start(self):
        # FIXME (1) later : may replace BeamOn with DoEventLoop
        # to allow better control on geometry between the different runs
        # (2) : check estimated nb of particle, warning if too large
        # start the master thread (only main thread)
        self.g4_master_source_manager.StartMasterThread()

        # once terminated, packup the sources (if needed)
        for source in self.sources:
            source.prepare_output()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# import pdb
import opengate as gate
import uproot
import sys
import pathlib
import os

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# multithread ?
argv = sys.argv
n = 1
if len(argv) > 1:
    n = int(argv[1])

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = n
ui.random_seed = 12344321
print(ui)

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
nm = gate.g4_units("nm")
keV = gate.g4_units("keV")
Bq = gate.g4_units("Bq")
sec = gate.g4_units("s")

# set the world size like in the Gate macro
world = sim.world
world.size = [1 * m, 1 * m, 2 * m]

# waterbox (not really used here)
waterbox = sim.add_volume("Box", "waterbox")
waterbox.size = [10 * cm, 10 * cm, 10 * cm]
waterbox.translation = [0 * cm, 0 * cm, 0 * cm]
waterbox.material = "G4_AIR"

# detector
detector = sim.add_volume("Box", "detector")
detector.size = [80 * cm, 80 * cm, 1 * nm]
detector.translation = [0, 0, 30 * cm]
detector.material = "G4_BGO"
detector.color = [1, 0, 0, 1]

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMZ"
p.enable_decay = True
cuts = p.production_cuts
cuts.world.gamma = 1 * mm
cuts.world.proton = 1 * mm
cuts.world.electron = 1 * mm
cuts.world.positron = 1 * mm

# source #1
source1 = sim.add_source("GenericSource", "source1")
source1.particle = "gamma"
source1.energy.mono = 100 * keV
source1.position.type = "disc"
source1.position.radius = 2 * cm
source1.position.translation = [0, 0, -10 * cm]
source1.direction.type = "focused"
source1.direction.focus_point = [0, 0, 0]
source1.activity = 10000 * Bq / ui.number_of_threads
source1.half_life = 2 * sec

# source #2
source2 = sim.add_source("GenericSource", "source2")
source2.particle = "gamma"
source2.energy.mono = 200 * keV
source2.position.type = "disc"
source2.position.radius = 2 * cm
source2.position.translation = [0, 0, -10 * cm]
source2.direction.type = "focused"
source2.direction.focus_point = [0, 0, 0]
source2.activity = 10000 * Bq / ui.number_of_threads
# source2.n = 50

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# hit actor
ta = sim.add_actor("PhaseSpaceActor", "PhaseSpace")
ta.mother = "detector"
ta.attributes = ["KineticEnergy", "GlobalTime"]
ta.output = paths.output / "test022_half_life.root"

# timing
sim.run_timing_intervals = [
    [1 * sec, 10 * sec],
    [15 * sec, 20 * sec],
]  # "hole" in the timeline

# start simulation
output = sim.start()

# get result
stats = output.get_actor("Stats")
print(stats)

# read phsp

root = uproot.open(ta.output)
branch = root["PhaseSpace"]["GlobalTime"]
time = branch.array(library="numpy") / sec
branch = root["PhaseSpace"]["KineticEnergy"]
E = branch.array(library="numpy")

# consider time of arrival for both sources
time1 = time[E < 110 * keV]
time2 = time[E > 110 * keV]

# fit for half life
start_time = sim.run_timing_intervals[0][0] / sec
end_time = sim.run_timing_intervals[0][1] / sec
hl, xx, yy = gate.fit_exponential_decay(time1, start_time, end_time)
# compare with source half_life (convert in sec)
tol = 0.05
hl_ref = source1.half_life / sec
diff = abs(hl - hl_ref) / hl_ref
is_ok = b = diff < tol
diff *= 100
gate.print_test(b, f"Half life {hl_ref:.2f} sec vs {hl:.2f} sec : {diff:.2f}% ")

# check second source
m = len(time2)
start_time2 = sim.run_timing_intervals[1][0] / sec
end_time2 = sim.run_timing_intervals[1][1] / sec
# number of elements is around activity times the duration (per thread)
m_ref = (
    source2.activity
    / Bq
    * (end_time - start_time + end_time2 - start_time2)
    * ui.number_of_threads
)
diff = abs(m - m_ref) / m_ref
b = diff < tol
diff *= 100
gate.print_test(b, f"Events for source #2:  {m_ref} vs {m} -> {diff:.2f}% ")
is_ok = is_ok and b

# check thread
b = ui.number_of_threads * len(sim.run_timing_intervals) == stats.counts.run_count
gate.print_test(b, f"Number of run: {stats.counts.run_count}")

is_ok = is_ok and b

# plot debug
import matplotlib.pyplot as plt

fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(5, 5))
a = ax
a.hist(
    time1,
    bins=100,
    label="decay source",
    histtype="stepfilled",
    alpha=0.5,
    density=True,
)
a.hist(
    time2,
    bins=100,
    label="constant source",
    histtype="stepfilled",
    alpha=0.5,
    density=True,
)
a.plot(xx, yy, label="fit half-life {:.2f} sec".format(hl))
a.legend()
a.set_xlabel("time (s)")
a.set_ylabel("detected photon")
# plt.show()

fn = paths.output / "test022_half_life_fit.png"
print("Figure in ", fn)
plt.savefig(fn)

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4
import numpy as np
from scipy.spatial.transform import Rotation
from box import Box

"""
A rotation matrix (3x3) can be represented by:
- G4RotationMatrix in module opengate_core
- np.array in module numpy
- Rotation in module scipy.spatial.transform

With scipy and np:
- rot_np = rot_scipy.as_matrix()
- rot_scipy = Rotation.from_matrix(rot_np)

With G4RotationMatrix
- rot_g4 = rot_np_as_g4(rot_np)
- rot_np = rot_g4_as_np(rot_g4)

Also for G4ThreeVector
- v_np = vec_g4_as_np(v_g4)
- v_g4 = vec_np_as_g4(v_np)
"""


def is_rotation_matrix(R):
    """
    https://stackoverflow.com/questions/53808503/how-to-test-if-a-matrix-is-a-rotation-matrix
    """
    # square matrix test
    if R.ndim != 2 or R.shape[0] != R.shape[1]:
        return False
    should_be_identity = np.allclose(R.dot(R.T), np.identity(R.shape[0], np.float_))
    should_be_one = np.allclose(np.linalg.det(R), 1)
    return should_be_identity and should_be_one


def vec_np_as_g4(v):
    return g4.G4ThreeVector(v[0], v[1], v[2])


def vec_g4_as_np(v):
    vnp = np.zeros(3)
    vnp[0] = v.x
    vnp[1] = v.y
    vnp[2] = v.z
    return vnp


def rot_np_as_g4(rot):
    if not is_rotation_matrix(rot):
        gate.fatal(f"This matrix is not a rotation matrix (not orthogonal): \n{rot}")
    try:
        r = g4.HepRep3x3(
            rot[0, 0],
            rot[0, 1],
            rot[0, 2],
            rot[1, 0],
            rot[1, 1],
            rot[1, 2],
            rot[2, 0],
            rot[2, 1],
            rot[2, 2],
        )
    except Exception as e:
        s = f"Cannot convert the rotation {rot} to a 3x3 matrix. Exception is: "
        s += str(e)
        gate.fatal(s)
    a = g4.G4RotationMatrix()
    a.set(r)
    return a


def rot_g4_as_np(rot):
    r = np.zeros(shape=(3, 3))
    r[0, 0] = rot.xx()
    r[0, 1] = rot.xy()
    r[0, 2] = rot.xz()
    r[1, 0] = rot.yx()
    r[1, 1] = rot.yy()
    r[1, 2] = rot.yz()
    r[2, 0] = rot.zx()
    r[2, 1] = rot.zy()
    r[2, 2] = rot.zz()
    if not is_rotation_matrix(r):
        gate.fatal(f"The G4 matrix is not a rotation matrix (not orthogonal): \n{rot}")
    return r


def get_vol_g4_translation(vol):
    # the input can be a class UserInfo or a Box
    if isinstance(vol, gate.UserInfo):
        vd = vol.__dict__
    else:
        vd = vol
    if "translation" not in vd:
        gate.fatal(f'Cannot find the key "translation" into this volume: {vol}')
    try:
        t = vec_np_as_g4(vol.translation)
        return t
    except Exception as e:
        s = f"Cannot convert the translation {vol.translation} to a 3D vector. Exception is: "
        s += str(e)
        gate.fatal(s)


def get_vol_g4_rotation(vol):
    # the input can be a class UserInfo or a Box
    if isinstance(vol, gate.UserInfo):
        vd = vol.__dict__
    else:
        vd = vol
    if "rotation" not in vd:
        gate.fatal(f'Cannot find the key "rotation" into this volume: {vol}')
    return rot_np_as_g4(vol.rotation)


def get_vol_g4_transform(vol):
    translation = get_vol_g4_translation(vol)
    rotation = get_vol_g4_rotation(vol)
    return g4.G4Transform3D(rotation, translation)


def get_translation_from_rotation_with_center(rot, center):
    center = np.array(center)
    t = rot.apply(-center) + center
    # note: apply is the same than rot.as_matrix().dot()
    return t


def get_transform_orbiting(position, axis, angle_deg):
    p = np.array(position)
    rot = Rotation.from_euler(axis, angle_deg, degrees=True)
    t = rot.apply(p)
    return t, rot.as_matrix()


def get_transform_world_to_local(vol_name):
    # cumulated translation and rotation
    ctr = [0, 0, 0]
    crot = Rotation.identity().as_matrix()
    first = True
    while vol_name != gate.__world_name__:
        pv = g4.G4PhysicalVolumeStore.GetInstance().GetVolume(vol_name, False)
        tr = vec_g4_as_np(pv.GetObjectTranslation())
        rot = rot_g4_as_np(pv.GetObjectRotation())
        if first:
            ctr = tr
            crot = rot
            first = False
        else:
            crot = np.matmul(rot, crot)
            ctr = rot.dot(ctr) + tr
        vol_name = pv.GetMotherLogical().GetName()

    return ctr, crot


def repeat_ring(name, start_deg, nb, translation, axis=[0, 0, 1]):
    """
    Build a repeater for the given volume name, according to a ring rotation.
        start_deg *must* be in degrees
        nb is the number of repeated positions
        translation is the initial translation of the volume according to the center
        axis is the rotation axis
    The output is a dict (Box) of all positions (name + translation + rotation) than can be set
    to the 'repeat' member of a volume.
    """
    le = []
    step = np.pi * 2 / nb
    angle = np.deg2rad(start_deg)
    for i in range(nb):
        e = Box()
        e.name = f"{name}_{i}"
        r = Rotation.from_rotvec(angle * np.array(axis))
        e.rotation = r.as_matrix()
        e.translation = r.apply(translation)
        le.append(e)
        angle += step
    return le


def repeat_array(name, size, translation):
    start = [-(x - 1) * y / 2.0 for x, y in zip(size, translation)]
    return repeat_array_start(name, start, size, translation)


def repeat_array_start(name, start, size, translation):
    le = [
        {
            "name": f"{name}_{x}_{y}_{z}",
            "rotation": Rotation.identity().as_matrix(),
            "translation": [
                start[0] + translation[0] * x,
                start[1] + translation[1] * y,
                start[2] + translation[2] * z,
            ],
        }
        for x, y, z in np.ndindex(size[0], size[1], size[2])
    ]
    return le


def build_param_repeater(
    sim, mother_name, repeated_vol_name, size, translation, rot=None
):
    vol = sim.get_volume_user_info(repeated_vol_name)
    vol.build_physical_volume = False
    param = sim.add_volume("RepeatParametrised", f"{repeated_vol_name}_param")
    param.mother = mother_name
    param.repeated_volume_name = repeated_vol_name
    param.rotation = rot
    param.linear_repeat = size
    param.translation = translation
    param.start = [-(x - 1) * y / 2.0 for x, y in zip(size, translation)]
    param.offset_nb = 1
    param.offset = [0, 0, 0]
    return param


def volume_orbiting_transform(axis, start, end, n, initial_t, initial_rot):
    angle = start
    step_angle = (end - start) / n
    translations = []
    rotations = []
    for r in range(n):
        irot = Rotation.from_matrix(initial_rot)
        t, rot = gate.get_transform_orbiting(initial_t, axis, angle)
        rot = Rotation.from_matrix(rot)
        rot = rot * irot
        translations.append(t)
        r = gate.rot_np_as_g4(rot.as_matrix())
        rotations.append(r)
        angle += step_angle
    return translations, rotations



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.actor.helpers_actor
import opengate.contrib.spect_ge_nm670 as gate_spect
from box import Box
import json
import numpy as np

import opengate.physics.helpers_physics

paths = gate.get_default_test_paths(__file__, "")

radionuclides = ["Tc99m", "Lu177", "In111", "I131"]
ref_collimators = ["lehr", "megp", "megp", "hegp"]

# Test 1
is_ok = True
for rad, col in zip(radionuclides, ref_collimators):
    collimator = gate_spect.get_collimator(rad)
    ok = True
    if collimator != col:
        ok = False
    gate.print_test(ok, f"Rad {rad}: the collimator is {col}")
    is_ok = is_ok and ok

# Test 2
print()
sim = gate.Simulation()
digit = Box()
digit_ns = Box()
for rad in radionuclides:
    channels = gate.get_simplified_digitizer_channels_rad("fake_spect", rad, True)
    # cc = gate_spect.add_digitizer_energy_windows(sim, 'fake_crystal', channels)
    digit[rad] = channels
    channels = gate.get_simplified_digitizer_channels_rad("fake_spect", rad, False)
    # cc = gate_spect.add_digitizer_energy_windows(sim, 'fake_crystal', channels)
    digit_ns[rad] = channels

print(digit)
print()
print(digit_ns)

# reference
"""outfile = open(paths.output_ref / "t046_digitizer.json", "w")
json.dump(digit, outfile, indent=4)
outfile = open(paths.output_ref / "t046_digitizer_ns.json", "w")
json.dump(digit_ns, outfile, indent=4)"""

# check
ref_digit = json.loads(open(paths.output_ref / "t046_digitizer.json").read())
ref_digit_ns = json.loads(open(paths.output_ref / "t046_digitizer_ns.json").read())
ok = digit == ref_digit
gate.print_test(ok, f"Test channels (with scatter): {ok}")
is_ok = is_ok and ok
ok = digit_ns == ref_digit_ns
gate.print_test(ok, f"Test channels (without scatter): {ok}")
is_ok = is_ok and ok

# Test 3
print()
yields = [0.885, 0.172168, 1.847315, 1.0024600000000004]
i = 0
for rad in radionuclides:
    w, e = gate.get_rad_gamma_energy_spectrum(rad)
    tw = np.array(w).sum()
    ok = tw == yields[i]
    gate.print_test(ok, f"Test yield {rad}: {tw} {yields[i]} {ok}")
    is_ok = is_ok and ok
    i += 1

# end
gate.test_ok(is_ok)



import opengate as gate


class ActorManager:
    """
    Manage all the actors in the simulation
    """

    def __init__(self, simulation):
        self.simulation = simulation
        self.user_info_actors = {}

    def __str__(self):
        v = [v.name for v in self.user_info_actors.values()]
        s = f'{" ".join(v)} ({len(self.user_info_actors)})'
        return s

    def __del__(self):
        # print("del ActorManager")
        pass

    def __getstate__(self):
        # needed to not pickle. Need to reset user_info_actors to avoid to store the actors
        self.user_info_actors = {}
        return self.__dict__

    def dump(self):
        n = len(self.user_info_actors)
        s = f"Number of Actors: {n}"
        for actor in self.user_info_actors.values():
            if n > 1:
                a = "\n" + "-" * 20
            else:
                a = ""
            a += f"\n {actor}"
            s += gate.indent(2, a)
        return s

    def get_actor_user_info(self, name):
        if name not in self.user_info_actors:
            gate.fatal(
                f"The actor {name} is not in the current "
                f"list of actors: {self.user_info_actors}"
            )
        return self.user_info_actors[name]

    def add_actor(self, actor_type, name):
        # check that another element with the same name does not already exist
        gate.assert_unique_element_name(self.user_info_actors, name)
        # build it
        a = gate.UserInfo("Actor", actor_type, name)
        # append to the list
        self.user_info_actors[name] = a
        # return the info
        return a



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from test013_phys_lists_helpers import create_pl_sim, phys_em_parameters

paths = gate.get_default_test_paths(__file__, "gate_test013_phys_lists")

# create simulation
sim = create_pl_sim()

# remove ion sources
sim.source_manager.user_info_sources.pop("ion1")
sim.source_manager.user_info_sources.pop("ion2")

# change physics
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMZ"

# em parameters
phys_em_parameters(p)

print("Phys list cuts:")
print(sim.physics_manager.dump_cuts())

# start simulation
# sim.set_g4_verbose(True)
# sim.apply_g4_command("/tracking/verbose 1")
output = sim.start()

stats = output.get_actor("Stats")

# Gate mac/main_2.mac
stats_ref = gate.read_stat_file(paths.gate_output / "stat_2.txt")
is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.07)

gate.test_ok(is_ok)



from .GenericSource import *
from .GANSourceDefaultGenerator import GANSourceDefaultGenerator
import time


class GANSourceConditionalGenerator(GANSourceDefaultGenerator):
    def __init__(self, user_info, generate_condition_function):
        super().__init__(user_info)
        self.generate_condition = generate_condition_function

    def __getstate__(self):
        super().__getstate__()
        # we cannot pickle this function (not fully sure why)
        self.generate_condition = None
        return self.__dict__

    def generate_condition(self, n):
        gate.fatal(
            f'Error: to use GANSourceConditionalGenerator,  you must provide a function "f" '
            f'that take a single int "n" as input and generate n condition samples. '
            f'This function "f" must be set with generator.generate_condition = f'
        )
        return None

    def generator(self, source):
        """
        Generate particles with a GAN, considering conditional vectors.
        """

        # get the info
        g = self.gan_info
        n = self.user_info.batch_size
        start = None

        # verbose and timing ?
        if self.user_info.verbose_generator:
            start = time.time()
            print(f"Generate {n} particles from GAN ", end="")

        # generate cond
        cond = self.generate_condition(n)

        # generate samples (this is the most time-consuming part)
        if self.user_info.cond_debug:
            # debug : do not run GAN, only consider the conditions
            # needed by test 047
            fake = cond
        else:
            fake = self.gaga.generate_samples2(
                g.params,
                g.G,
                g.D,
                n=n,
                batch_size=n,
                normalize=False,
                to_numpy=True,
                cond=cond,
                silence=True,
            )

        # consider the names of the output keys position/direction/energy/time/weight
        self.get_output_keys()

        # if debug, the GAN is not used.
        if self.user_info.cond_debug:
            g.position_gan_index = [0, 1, 2]
            g.direction_gan_index = [3, 4, 5]
            g.energy_is_set_by_GAN = False
            g.energy_type = False

        # move particle backward ?
        self.move_backward(g, fake)

        # copy to cpp (g4)
        self.copy_generated_particle_to_g4(source, g, fake)

        # verbose
        if self.user_info.verbose_generator:
            end = time.time()
            print(f"in {end - start:0.2f} sec (GPU={g.params.current_gpu})")



import itk
import numpy as np
from box import Box
import opengate as gate
import opengate_core as g4
from scipy.spatial.transform import Rotation


def update_image_py_to_cpp(py_img, cpp_img, copy_data=False):
    cpp_img.set_size(py_img.GetLargestPossibleRegion().GetSize())
    cpp_img.set_spacing(py_img.GetSpacing())
    cpp_img.set_origin(py_img.GetOrigin())
    # this is needed !
    cpp_img.set_region(
        py_img.GetLargestPossibleRegion().GetIndex(),
        py_img.GetLargestPossibleRegion().GetSize(),
    )
    # It is really a pain to convert GetDirection into
    # something that can be read by SetDirection !
    d = py_img.GetDirection().GetVnlMatrix().as_matrix()
    rotation = itk.GetArrayFromVnlMatrix(d)
    cpp_img.set_direction(rotation)
    if copy_data:
        arr = itk.array_view_from_image(py_img)
        cpp_img.from_pyarray(arr)


def itk_dir_to_rotation(dir):
    return itk.GetArrayFromVnlMatrix(dir.GetVnlMatrix().as_matrix())


def create_3d_image(size, spacing, pixel_type="float", allocate=True, fill_value=0):
    dim = 3
    pixel_type = itk.ctype(pixel_type)
    image_type = itk.Image[pixel_type, dim]
    img = image_type.New()
    region = itk.ImageRegion[dim]()
    size = np.array(size)
    region.SetSize(size.tolist())
    region.SetIndex([0, 0, 0])
    spacing = np.array(spacing)
    img.SetRegions(region)
    img.SetSpacing(spacing)
    # (default origin and direction)
    if allocate:
        img.Allocate()
        img.FillBuffer(fill_value)
    return img


def create_image_like(like_image, allocate=True, pixel_type=""):
    # TODO fix pixel_type -> copy from image rather than argument
    info = get_info_from_image(like_image)

    if pixel_type:
        img = create_3d_image(
            info.size, info.spacing, pixel_type=pixel_type, allocate=allocate
        )
    else:
        img = create_3d_image(info.size, info.spacing, allocate=allocate)
    img.SetOrigin(info.origin)
    img.SetDirection(info.dir)
    return img


def create_image_like_info(info, allocate=True):
    img = create_3d_image(info.size, info.spacing, allocate=allocate)
    img.SetOrigin(info.origin)
    img.SetDirection(info.dir)
    return img


def get_info_from_image(image):
    info = Box()
    info.size = np.array(itk.size(image)).astype(int)
    info.spacing = np.array(image.GetSpacing())
    info.origin = np.array(image.GetOrigin())
    info.dir = image.GetDirection()
    return info


def read_image_info(filename):
    filename = str(filename)
    image_IO = itk.ImageIOFactory.CreateImageIO(
        filename, itk.CommonEnums.IOFileMode_ReadMode
    )
    if not image_IO:
        gate.fatal(f"Cannot read the header of this image file (itk): {filename}")
    image_IO.SetFileName(filename)
    image_IO.ReadImageInformation()
    info = Box()
    info.filename = filename
    n = info.size = image_IO.GetNumberOfDimensions()
    info.size = np.ones(n).astype(int)
    info.spacing = np.ones(n)
    info.origin = np.ones(n)
    info.dir = np.ones((n, n))
    for i in range(n):
        info.size[i] = image_IO.GetDimensions(i)
        info.spacing[i] = image_IO.GetSpacing(i)
        info.origin[i] = image_IO.GetOrigin(i)
        info.dir[i] = image_IO.GetDirection(i)
    return info


def get_translation_between_images_center(img_name1, img_name2):
    """
    The two images are considered in the same physical space (coordinate system).
    This function computes the translation between their centers.
    Warning, the ITK image origin consider the center of the first voxel, we thus
    consider half a pixel shift for the center.
    """
    info1 = read_image_info(img_name1)
    info2 = read_image_info(img_name2)
    # get the center of the first image in img coordinate system
    center1 = info1.size / 2.0 * info1.spacing + info1.origin - info1.spacing / 2.0
    # get the center of the second image in img coordinate system
    center2 = info2.size / 2.0 * info2.spacing + info2.origin - info2.spacing / 2.0
    return center2 - center1


def get_origin_wrt_images_g4_position(img_info1, img_info2, translation):
    """
    The two images are considered in the same GATE physical space (coordinate system), so according to the
    centers of both images (+translation).
    This function computes the origin for the second image such as the two images will be in the same
    physical space of the first image.
    Warning, the ITK image origin considers the center of the first voxel, we thus
    consider half a pixel shift for the center.
    """
    half_size1 = img_info1.size * img_info1.spacing / 2.0
    half_size2 = img_info2.size * img_info2.spacing / 2.0
    origin = (
        img_info1.origin
        + half_size1
        - half_size2
        + translation
        - img_info1.spacing / 2.0
        + img_info2.spacing / 2
    )
    return origin


def get_cpp_image(cpp_image):
    arr = cpp_image.to_pyarray()
    image = itk_image_view_from_array(arr)
    image.SetOrigin(cpp_image.origin())
    image.SetSpacing(cpp_image.spacing())
    return image


def itk_image_view_from_array(arr):
    """
    When the input numpy array is of shape [1,1,x], the conversion to itk image fails:
    the output image size is with the wrong dimensions.
    We thus 'patch' itk.image_view_from_array to correct the size.

    Not fully sure if this is the way to go.
    """
    image = itk.image_view_from_array(arr)
    if len(arr.shape) == 3 and arr.shape[1] == arr.shape[2] == 1:
        new_region = itk.ImageRegion[3]()
        new_region.SetSize([1, 1, arr.shape[0]])
        image.SetRegions(new_region)
        image.Update()
    return image


def get_image_center(image):
    info = read_image_info(image)
    center = info.size * info.spacing / 2.0  # + info.spacing / 2.0
    return center


def get_translation_from_iso_center(img_info, rot, iso_center, centered):
    if centered:
        # cf Gate GateVImageVolume.cc, function UpdatePositionWithIsoCenter
        iso_center = iso_center - img_info.origin
        center = img_info.size * img_info.spacing / 2.0
        iso_center -= center
        t = rot.apply(iso_center)
        return t
    gate.fatal(f"not implemented yet")


def get_physical_volume(volume_engine, vol_name, physical_volume_index):
    vol = volume_engine.get_volume(vol_name)
    vols = vol.g4_physical_volumes
    if len(vols) == 0:
        gate.fatal(
            f'The function "attach_image_to_volume" can only be used after initialization'
        )
    if physical_volume_index is None and len(vols) > 1:
        gate.fatal(
            f"There are {len(vols)} physical volumes attached to the {vol_name}, "
            f'"physical_volume_index" must be set explicitly.'
        )
    if physical_volume_index is not None and len(vols) <= physical_volume_index:
        gate.fatal(
            f"Cannot find phys vol {physical_volume_index}, in the list of physical "
            f"volumes of {vol_name} ({len(vols)})"
        )
    if physical_volume_index is None:
        return vols[0]
    return vols[physical_volume_index]


def attach_image_to_physical_volume(
    phys_vol_name, image, initial_translation=None, initial_rotation=Rotation.identity()
):
    if initial_translation is None:
        initial_translation = [0, 0, 0]
    # FIXME rotation not implemented yet
    # get transform from world
    translation, rotation = gate.get_transform_world_to_local(phys_vol_name)
    # compute origin
    info = get_info_from_image(image)
    origin = -info.size * info.spacing / 2.0 + info.spacing / 2.0 + initial_translation
    origin = Rotation.from_matrix(rotation).apply(origin) + translation
    # set origin and direction
    image.SetOrigin(origin)
    image.SetDirection(rotation)


def create_image_with_volume_extent(sim, vol_name, spacing=[1, 1, 1], margin=0):
    pMin, pMax = gate.get_volume_bounding_limits(sim, vol_name)
    pMin = gate.vec_g4_as_np(pMin)
    pMax = gate.vec_g4_as_np(pMax)

    # define the new size and spacing
    spacing = np.array(spacing).astype(float)
    size = np.ceil((pMax - pMin) / spacing).astype(int)
    size = size + margin * 2

    # create image
    image = gate.create_3d_image(size, spacing)

    # the origin is considered at the center of first pixel
    # is it set such as the image is at the exact extent (bounding volume)
    # (the volume contour thus goes through the center of the first pixel)
    origin = pMin + spacing / 2.0 - margin
    image.SetOrigin(origin)
    return image


def voxelize_volume(se, vol_name, image):
    # simulation engine
    if not se.is_initialized:
        se.initialize()
    # initialization is needed because it builds the hierarchy of G4 volumes
    # that are needed by the "voxelize" function
    # get physical volume
    vol = se.volume_engine.get_volume(vol_name).g4_physical_volume
    if vol.GetMultiplicity() != 1:
        gate.warning(
            f"Warning the volume {vol_name} is multiple: "
            f"{vol.GetMultiplicity()}. Only first is considered"
        )

    # world volume
    world = se.volume_engine.get_volume("world").g4_physical_volume

    # navigator
    nav = g4.G4Navigator()
    nav.SetWorldVolume(world)

    # list of volume label
    labels = {}
    vox = g4.GateVolumeVoxelizer()
    gate.update_image_py_to_cpp(image, vox.fImage, False)
    vox.Voxelize(vol_name)

    image = gate.get_cpp_image(vox.fImage)
    labels = vox.fLabels
    return labels, image


def transform_images_point(p, img1, img2):
    index = img1.TransformPhysicalPointToIndex(p)
    pbis = img2.TransformIndexToPhysicalPoint(index)
    return [i for i in pbis]


def compute_image_3D_CDF(image):
    """
    Compute the three CDF (Cumulative Density Function) for the given image
    Warning; numpy order is ZYX

    :param image: itk image
    """
    # consider image as np array
    array = itk.array_view_from_image(image)

    # normalize
    array = array / np.sum(array)

    # Sum image on a single plane along X axis
    sumx = np.sum(array, axis=2)
    # Y axis, sum plane on a single axis along Y axis
    sumxy = np.sum(sumx, axis=1)
    # X 3D CDF
    cdf_x = []
    for i in range(array.shape[0]):  # Z
        cdf_x.append([])
        for j in range(array.shape[1]):  # Y
            # cumulated sum along X axis
            t = np.cumsum(array[i][j])
            # normalise if last value (sum) is not zero
            if t[-1] != 0:
                t = t / t[-1]
            cdf_x[i].append(t)

    # Y 2D CDF
    cdf_y = []
    for i in range(len(sumx)):  # Z
        t = np.cumsum(sumx[i])
        if t[-1] != 0:
            t = t / t[-1]
        cdf_y.append(t)

    # Z 1D CDF
    cdf_z = np.cumsum(sumxy) / np.sum(sumxy)

    # return
    return cdf_x, cdf_y, cdf_z


def scale_itk_image(img, scale):
    imgarr = itk.array_view_from_image(img)
    imgarr = imgarr * scale
    img2 = itk.image_from_array(imgarr)
    img2.CopyInformation(img)
    return img2


def divide_itk_images(
    img1_numerator, img2_denominator, filterVal=0, replaceFilteredVal=0
):
    imgarr1 = itk.array_view_from_image(img1_numerator)
    imgarr2 = itk.array_view_from_image(img2_denominator)
    imgarrOut = imgarr1.copy()
    L_filterInv = imgarr2 != filterVal
    imgarrOut[L_filterInv] = np.divide(imgarr1[L_filterInv], imgarr2[L_filterInv])

    imgarrOut[np.invert(L_filterInv)] = replaceFilteredVal
    imgarrOut = itk.image_from_array(imgarrOut)
    imgarrOut.CopyInformation(img1_numerator)
    return imgarrOut


def split_spect_projections(input_filenames, nb_ene):
    """
    The inputs are filenames of several images containing projections for a given spect head
    Each image is composed of nb_ene energy windows and XX angles.
    The number of angles is found by looking at the number of slices.

    The function computes nb_ene itk image with all angles and all heads merged into a list of
    projections stored as a 3D image, to make it easy to reconstruct with RTK.

    """
    nb_heads = len(input_filenames)

    # read the first image to get information
    img = itk.imread(str(input_filenames[0]))
    info = gate.get_info_from_image(img)
    imga = itk.array_view_from_image(img)

    nb_runs = imga.shape[0] // nb_ene
    nb_angles = nb_heads * nb_runs
    # print('Number of heads', nb_heads)
    # print('Number of E windows', nb_ene)
    # print('Number of run', nb_runs)
    # print('Number of angles', nb_angles)

    # create and allocate final images
    outputs_img = []
    outputs_arr = []
    size = [imga.shape[1], imga.shape[2], nb_angles]
    spacing = info.spacing
    for e in range(nb_ene):
        image = gate.create_3d_image(size, spacing)
        image.SetOrigin(img.GetOrigin())
        outputs_img.append(image)
        outputs_arr.append(itk.array_view_from_image(image))

    # loop on heads and create images
    s2 = 0
    for head in range(nb_heads):
        img = itk.imread(str(input_filenames[head]))
        imga = itk.array_view_from_image(img)
        e = 0
        for s in range(imga.shape[0]):
            outputs_arr[e][s2] = imga[s]
            e += 1
            if e >= nb_ene:
                e = 0
                s2 += 1

    # end
    return outputs_img



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test043_garf_helpers import *
import opengate.contrib.spect_ge_nm670 as gate_spect
import itk
import numpy as np

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = 123654
ui.number_of_threads = 1
colli = "lehr"

# units
activity = 1e6 * Bq / ui.number_of_threads

# world
sim_set_world(sim)

# spect head
spect, crystal = gate_spect.add_ge_nm67_spect_head(
    sim, "spect", collimator_type=colli, debug=ui.visu
)
spect_translation = 15 * cm
spect.translation = [0, 0, -spect_translation]
crystal_name = f"{spect.name}_crystal"

# physics
sim_phys(sim)

# sources
sim_source_test(sim, activity)

# digitizer
channels = [
    {"name": f"spectrum_{spect.name}", "min": 0 * keV, "max": 160 * keV},
    {"name": f"scatter_{spect.name}", "min": 114 * keV, "max": 126 * keV},
    {"name": f"peak140_{spect.name}", "min": 126 * keV, "max": 154 * keV},
]
proj = gate_spect.add_digitizer(sim, crystal_name, channels)
proj.spacing = [4.41806 * mm, 4.41806 * mm]
proj.output = paths.output / "test043_projection_analog.mhd"

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True

# start simulation
output = sim.start()

# print results at the end
stat = output.get_actor("stats")
print(stat)

# dump the output image with offset like in old gate (for comparison)
print("We change the spacing/origin to be compared to the old gate")
proj = output.get_actor(f"Projection_{crystal_name}")
spacing = np.array([4.41806 * mm, 4.41806 * mm, 1])
img = itk.imread(str(proj.user_info.output))
img.SetSpacing(spacing)
img.SetOrigin(spacing / 2.0)
fn = str(proj.user_info.output).replace(".mhd", "_offset.mhd")
itk.imwrite(img, fn)

# ----------------------------------------------------------------------------------------------------------------
# tests
print()
gate.warning("Tests stats file")
stats_ref = gate.read_stat_file(paths.gate_output / "stats_analog.txt")
stat.counts.run_count = 1  # force to one run (ref only have 1 thread)
is_ok = gate.assert_stats(stat, stats_ref, 0.01)

print()
gate.warning("Tests projection (old gate)")
is_ok = (
    gate.assert_images(
        paths.gate_output / "projection_analog.mhd",
        fn,
        stat,
        tolerance=75,
        ignore_value=0,
        axis="x",
    )
    and is_ok
)

print()
gate.warning("Tests projection (new)")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test043_projection_analog.mhd",
        proj.user_info.output,
        stat,
        tolerance=80,
        ignore_value=0,
        axis="x",
    )
    and is_ok
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.spect_ge_nm670 as gate_spect

paths = gate.get_default_test_paths(__file__, "")


def create_test(sim, nb_thread=1):
    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.running_verbose_level = gate.RUN
    ui.check_volumes_overlap = False
    ui.random_seed = 123456

    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    sec = gate.g4_units("second")
    deg = gate.g4_units("deg")
    kBq = 1000 * Bq
    MBq = 1000 * kBq

    """ ================================================== """
    # main parameters
    ui.visu = False
    ui.g4_verbose = False
    ui.visu_verbose = False
    ui.number_of_threads = nb_thread
    ac = 3 * MBq
    distance = 15 * cm
    psd = 6.11 * cm
    p = [0, 0, -(distance + psd)]
    """ ================================================== """

    # world size
    world = sim.world
    world.size = [1.5 * m, 1.5 * m, 1.5 * m]
    world.material = "G4_AIR"

    # spect head (debug mode = very small collimator)
    spect1, crystal = gate_spect.add_ge_nm67_spect_head(
        sim, "spect1", collimator_type="lehr", debug=ui.visu
    )
    spect1.translation, spect1.rotation = gate.get_transform_orbiting(p, "x", 180)

    # spect head (debug mode = very small collimator)
    spect2, crystal = gate_spect.add_ge_nm67_spect_head(
        sim, "spect2", collimator_type="lehr", debug=ui.visu
    )
    spect2.translation, spect2.rotation = gate.get_transform_orbiting(p, "x", 0)

    # physic list
    sim.set_cut("world", "all", 10 * mm)
    # sim.set_cut('spect1_crystal', 'all', 1 * mm)
    # sim.set_cut('spect2_crystal', 'all', 1 * mm)

    # source #1
    sources = []
    source = sim.add_source("GenericSource", "source1")
    source.particle = "gamma"
    source.energy.type = "mono"
    source.energy.mono = 140.5 * keV
    source.position.type = "sphere"
    source.position.radius = 2 * mm
    source.position.translation = [0, 0, 20 * mm]
    source.direction.type = "iso"
    source.direction.acceptance_angle.volumes = ["spect2", "spect1"]
    source.direction.acceptance_angle.intersection_flag = True
    source.direction.acceptance_angle.normal_flag = True
    source.direction.acceptance_angle.normal_vector = [0, 0, -1]
    source.direction.acceptance_angle.normal_tolerance = 10 * deg
    source.direction.acceptance_angle.skip_policy = "ZeroEnergy"
    source.activity = ac / ui.number_of_threads
    sources.append(source)

    # source #1
    source2 = sim.add_source("GenericSource", "source2")
    gate.copy_user_info(source, source2)
    source2.position.radius = 1 * mm
    source2.position.translation = [20 * mm, 0, -20 * mm]
    sources.append(source2)

    # add stat actor
    stat = sim.add_actor("SimulationStatisticsActor", "Stats")
    stat.output = paths.output / "test033_stats.txt"

    # add default digitizer (it is easy to change parameters if needed)
    proj = gate_spect.add_simplified_digitizer_Tc99m(
        sim, "spect1_crystal", paths.output / "test033_proj_1.mhd"
    )
    proj.origin_as_image_center = False
    proj = gate_spect.add_simplified_digitizer_Tc99m(
        sim, "spect2_crystal", paths.output / "test033_proj_2.mhd"
    )
    proj.origin_as_image_center = False

    # motion of the spect, create also the run time interval
    heads = [spect1, spect2]

    # create a list of run (total = 1 second)
    n = 10
    sim.run_timing_intervals = gate.range_timing(0, 1 * sec, n)

    for head in heads:
        motion = sim.add_actor("MotionVolumeActor", f"Move_{head.name}")
        motion.mother = head.name
        motion.translations, motion.rotations = gate.volume_orbiting_transform(
            "x", 0, 180, n, head.translation, head.rotation
        )
        motion.priority = 5

    return sources


def evaluate_test(output, sources, itol, ref_skipped):
    stats = output.get_actor("Stats")
    print(stats)
    # ref with _noaa
    # stats.write(paths.output_ref / "test033_stats.txt")

    se = 0
    ze = 0
    for source in sources:
        se += gate.get_source_skipped_events(output, source.name)
        ze += gate.get_source_zero_events(output, source.name)
    print(f"Skipped particles {se}")
    print(f"Zeros E particles {ze}")
    s = max(se, ze)

    # check nb of avoided events (either skipped or energy zero)
    gate.warning(f"Check skipped")
    tol = 0.01
    if ref_skipped != 0:
        d = abs(ref_skipped - s) / ref_skipped
    else:
        d = 0
    is_ok = d < tol
    gate.print_test(
        is_ok,
        f"Skipped particles ref={ref_skipped}, get {s} -> {d * 100}% vs tol={tol * 100}%",
    )

    # check stats
    gate.warning(f"Check stats")
    stats_ref = gate.read_stat_file(paths.output_ref / "test033_stats.txt")
    print(f"Steps counts not compared (was {stats.counts.step_count})")
    nbt = output.simulation.user_info.number_of_threads
    stats.counts.step_count = stats_ref.counts.step_count
    stats_ref.counts.run_count *= nbt
    if se > 0:
        print(f"Track counts not compared (was {stats.counts.track_count})")
        stats.counts.event_count += se
        print(f"Modify Events + skipped {stats.counts.event_count+se})")
        stats.counts.track_count = stats_ref.counts.track_count
    """if ze > 0:
        print(f"Track counts not compared (was {stats.counts.track_count})")
        stats.counts.event_count += ze
        print(f"Modify Events - ZE {stats.counts.event_count})")
        stats.counts.track_count = stats_ref.counts.track_count"""
    is_ok = gate.assert_stats(stats, stats_ref, 0.01) and is_ok

    # compare edep map
    gate.warning(f"Check images")
    is_ok = (
        gate.assert_images(
            paths.output_ref / "test033_proj_1.mhd",
            paths.output / "test033_proj_1.mhd",
            stats,
            tolerance=65,
            axis="x",
            sum_tolerance=itol,
        )
        and is_ok
    )
    is_ok = (
        gate.assert_images(
            paths.output_ref / "test033_proj_2.mhd",
            paths.output / "test033_proj_2.mhd",
            stats,
            tolerance=65,
            axis="x",
            sum_tolerance=itol,
        )
        and is_ok
    )

    return is_ok



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import opengate as gate
import uproot

paths = gate.get_default_test_paths(__file__, "")

l = gate.all_beta_plus_radionuclides
# l = ['F18', 'Ga68', 'O15']
# l = ['F18']

# references (02/2022)
# http://www.lnhb.fr/nuclear-data/module-lara/
rad_yields = {
    "F18": 0.968600992766,
    "Ga68": 0.8883814158496728,
    "Zr89": 0.22799992881708,
    "Na22": 0.8972935562750121,
    "C11": 0.9974862363857401,
    "N13": 0.9981749688051,
    "O15": 0.9988401691350001,
    "Rb82": 0.95410853035736,
}

# Get one color for each rad
fig, ax1 = plt.subplots(1, 1, figsize=(20, 10))
cmap = plt.get_cmap("Dark2")
rad_color = {}
n = len(l)
i = 0
for rad in l:
    rad_color[rad] = cmap(i / n)
    i += 1

for rad in l:
    data = gate.read_beta_plus_spectra(rad)
    x = data[:, 0]  # energy E(keV)
    y = data[:, 1]  # proba  dNtot/dE b+
    # normalize taking into account the bins density
    dx = gate.compute_bins_density(x)
    s = (y * dx).sum()
    y = y / s
    ax1.plot(x, y, label=rad, color=rad_color[rad])

plt.xlabel("Energy KeV")
plt.ylabel("Probability")
ax1.legend()
plt.text(2200, 0.0023, "BetaShape")
plt.text(2200, 0.0020, "Mougeot, Phys Rev C 91, 055504 (2015)")
plt.text(2200, 0.0017, "http://www.lnhb.fr/nuclear-data/module-lara")

# units
cm = gate.g4_units("cm")
m = gate.g4_units("m")
Bq = gate.g4_units("Bq")

# simulation
sim = gate.Simulation()
sim.user_info.visu = False
sim.world.size = [1 * m, 1 * m, 1 * m]
sim.world.material = "G4_Galactic"


def add_box(i):
    b = sim.add_volume("Box", f"b{i}")
    b.size = [1 * cm, 1 * cm, 1 * cm]
    b.translation = [2 * i * cm, 0 * cm, 0 * cm]
    b.material = "G4_Galactic"


tol = 0.03


def add_source(rad):
    si = len(rads)
    add_box(si)
    source = sim.add_source("GenericSource", f"source_{rad}")
    source.mother = f"b{si}"
    source.particle = "e+"
    source.energy.type = f"{rad}"
    source.position.type = "point"
    source.direction.type = "iso"
    """
        WARNING
        with real simulation, the activity should be weighted by the total yield !
    """
    total_yield = gate.get_rad_yield(rad)
    source.activity = activity  # * total_yield  <--- this should be taken into account in real simulation
    yi = rad_yields[rad]
    t = (total_yield - yi) / yi < tol
    gate.print_test(t, f"Rad {rad} total yield = {total_yield} vs {yi} (tol is {tol})")

    phsp = sim.add_actor("PhaseSpaceActor", f"phsp_{rad}")
    phsp.mother = f"b{si}"
    phsp.attributes = ["TrackVertexKineticEnergy"]
    phsp.output = paths.output / f"test031_{rad}.root"
    f = sim.add_filter("ParticleFilter", "f")
    f.particle = "e+"
    phsp.filters.append(f)
    rads.append(rad)


rads = []
activity = 100000 * Bq
for r in l:
    add_source(r)

"""
add_source('F18_analytic')
rad_color['F18_analytic'] = rad_color['F18']
add_source('O15_analytic')
rad_color['O15_analytic'] = rad_color['O15']
add_source('C11_analytic')
rad_color['C11_analytic'] = rad_color['C11']
"""

s = sim.add_actor("SimulationStatisticsActor", "stats")
s.track_types_flag = True

output = sim.start()

# print results
stats = output.get_actor("stats")
print(stats)

# plot
for i in range(len(rads)):
    rad = rads[i]
    output = paths.output / f"test031_{rad}.root"
    data = uproot.open(output)[f"phsp_{rad}"]
    data = data.arrays(library="numpy")["TrackVertexKineticEnergy"] * 1000  # MeV to KeV
    ax1.hist(
        data,
        bins=200,
        density=True,
        histtype="stepfilled",
        alpha=0.5,
        label=f"{rads[i]}",
        color=rad_color[rad],
    )

f = paths.output / "test031.png"
ax1.legend(loc="upper center")
plt.savefig(f)
print(f"Figure save in {f}")

# compute diff
# ax2 = ax1.twinx()
is_ok = True
tol = 4
for rad in rads:
    # input
    output = paths.output_ref / f"test031_{rad}.root"
    data_ref = uproot.open(output)[f"phsp_{rad}"]
    data_ref = (
        data_ref.arrays(library="numpy")["TrackVertexKineticEnergy"] * 1000
    )  # MeV to KeV
    hist_ref, bins_ref = np.histogram(
        data_ref, range=(data_ref.min(), data_ref.max()), bins=100, density=False
    )
    """ax2.hist(data_ref, bins=100, density=False,
             range=(data_ref.min(), data_ref.max()), histtype='stepfilled',
             alpha=0.5, label=f'{rads[i]}', color='r')"""
    # output
    output = paths.output / f"test031_{rad}.root"
    data = uproot.open(output)[f"phsp_{rad}"]
    data = data.arrays(library="numpy")["TrackVertexKineticEnergy"] * 1000  # MeV to KeV
    hist, bins = np.histogram(
        data, range=(data_ref.min(), data_ref.max()), bins=100, density=False
    )
    """ax2.hist(data, bins=100, density=False,
             range=(data_ref.min(), data_ref.max()), histtype='stepfilled', alpha=0.5,
             label=f'{rads[i]}', color='b')"""
    # differences
    mean = hist_ref.sum() / len(hist_ref)
    msad = np.sum(np.abs(np.subtract(hist_ref, hist))) / len(hist_ref) / mean * 100
    t = msad < tol
    gate.print_test(t, f"Mean bin difference for {rad} is {msad:.2f}% (tol is {tol}%)")
    is_ok = is_ok and t

gate.test_ok(is_ok)



import opengate_core as g4
import itk
import numpy as np
import opengate as gate
from scipy.spatial.transform import Rotation


class DoseActor(g4.GateDoseActor, gate.ActorBase):
    """
    DoseActor: compute a 3D edep/dose map for deposited
    energy/absorbed dose in the attached volume

    The dose map is parameterized with:
        - size (number of voxels)
        - spacing (voxel size)
        - translation (according to the coordinate system of the "attachedTo" volume)
        - no rotation

    Position:
    - by default: centered according to the "attachedTo" volume center
    - if the attachedTo volume is an Image AND the option "img_coord_system" is True:
        the origin of the attachedTo image is used for the output dose.
        Hence, the dose can be superimposed with the attachedTo volume

    Options
        - edep only for the moment
        - later: add dose, uncertainty, squared etc

    """

    type_name = "DoseActor"

    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        # required user info, default values
        mm = gate.g4_units("mm")
        user_info.size = [10, 10, 10]
        user_info.spacing = [1 * mm, 1 * mm, 1 * mm]
        user_info.output = "edep.mhd"  # FIXME change to 'output' ?
        user_info.translation = [0, 0, 0]
        user_info.img_coord_system = None
        user_info.output_origin = None
        user_info.uncertainty = True
        user_info.gray = False
        user_info.physical_volume_index = None
        user_info.hit_type = "random"

    def __init__(self, user_info):
        gate.ActorBase.__init__(self, user_info)
        g4.GateDoseActor.__init__(self, user_info.__dict__)
        # attached physical volume (at init)
        self.g4_phys_vol = None
        # default image (py side)
        self.py_edep_image = None
        self.py_dose_image = None
        self.py_temp_image = None
        self.py_square_image = None
        self.py_last_id_image = None
        # default uncertainty
        self.uncertainty_image = None
        # internal states
        self.img_origin_during_run = None
        self.first_run = None
        self.output_origin = None

    def __str__(self):
        u = self.user_info
        s = f'DoseActor "{u.name}": dim={u.size} spacing={u.spacing} {u.output} tr={u.translation}'
        return s

    def __getstate__(self):
        # superclass getstate
        gate.ActorBase.__getstate__(self)
        # do not pickle itk images
        self.py_edep_image = None
        self.py_dose_image = None
        self.py_temp_image = None
        self.py_square_image = None
        self.py_last_id_image = None
        self.uncertainty_image = None
        return self.__dict__

    def initialize(self, volume_engine=None):
        """
        At the start of the run, the image is centered according to the coordinate system of
        the mother volume. This function computes the correct origin = center + translation.
        Note that there is a half-pixel shift to align according to the center of the pixel,
        like in ITK.
        """
        super().initialize(volume_engine)
        # create itk image (py side)
        size = np.array(self.user_info.size)
        spacing = np.array(self.user_info.spacing)
        self.py_edep_image = gate.create_3d_image(size, spacing)
        # compute the center, using translation and half pixel spacing
        self.img_origin_during_run = (
            -size * spacing / 2.0 + spacing / 2.0 + self.user_info.translation
        )
        # for initialization during the first run
        self.first_run = True

    def StartSimulationAction(self):
        # init the origin and direction according to the physical volume
        # (will be updated in the BeginOfRun)
        try:
            self.g4_phys_vol = gate.get_physical_volume(
                self.volume_engine,
                self.user_info.mother,
                self.user_info.physical_volume_index,
            )
        except:
            gate.fatal(f"Error in the DoseActor {self.user_info.name}")
        gate.attach_image_to_physical_volume(
            self.g4_phys_vol.GetName(), self.py_edep_image, self.user_info.translation
        )

        # Set the real physical volume name
        self.fPhysicalVolumeName = str(self.g4_phys_vol.GetName())

        # FIXME for multiple run and motion
        if not self.first_run:
            gate.warning(f"Not implemented yet: DoseActor with several runs")
        # send itk image to cpp side, copy data only the first run.
        gate.update_image_py_to_cpp(
            self.py_edep_image, self.cpp_edep_image, self.first_run
        )

        # for uncertainty
        if self.user_info.uncertainty:
            self.py_temp_image = gate.create_image_like(self.py_edep_image)
            self.py_square_image = gate.create_image_like(self.py_edep_image)
            self.py_last_id_image = gate.create_image_like(self.py_edep_image)
            gate.update_image_py_to_cpp(
                self.py_temp_image, self.cpp_temp_image, self.first_run
            )
            gate.update_image_py_to_cpp(
                self.py_square_image, self.cpp_square_image, self.first_run
            )
            gate.update_image_py_to_cpp(
                self.py_last_id_image, self.cpp_last_id_image, self.first_run
            )

        # for dose in Gray
        if self.user_info.gray:
            self.py_dose_image = gate.create_image_like(self.py_edep_image)
            gate.update_image_py_to_cpp(
                self.py_dose_image, self.cpp_dose_image, self.first_run
            )

        # now, indicate the next run will not be the first
        self.first_run = False

        # If attached to a voxelized volume, we may want to use its coord system.
        # So, we compute in advance what will be the final origin of the dose map
        vol_name = self.user_info.mother
        vol_type = self.simulation.get_volume_user_info(vol_name).type_name
        self.output_origin = self.img_origin_during_run

        # FIXME put out of the class ?
        if vol_type == "Image":
            if self.user_info.img_coord_system:
                vol = self.volume_engine.g4_volumes[vol_name]
                # Translate the output dose map so that its center correspond to the image center.
                # The origin is thus the center of the first voxel.
                img_info = gate.get_info_from_image(vol.image)
                dose_info = gate.get_info_from_image(self.py_edep_image)
                self.output_origin = gate.get_origin_wrt_images_g4_position(
                    img_info, dose_info, self.user_info.translation
                )
        else:
            if self.user_info.img_coord_system:
                gate.warning(
                    f'DoseActor "{self.user_info.name}" has '
                    f"the flag img_coord_system set to True, "
                    f"but it is not attached to an Image "
                    f'volume ("{vol_name}", of type "{vol_type}"). '
                    f"So the flag is ignored."
                )
        # user can set the output origin
        if self.user_info.output_origin is not None:
            if self.user_info.img_coord_system:
                gate.warning(
                    f'DoseActor "{self.user_info.name}" has '
                    f"the flag img_coord_system set to True, "
                    f"but output_origin is set, so img_coord_system ignored."
                )
            self.output_origin = self.user_info.output_origin

    def EndSimulationAction(self):
        g4.GateDoseActor.EndSimulationAction(self)

        # Get the itk image from the cpp side
        # Currently a copy. Maybe latter as_pyarray ?
        self.py_edep_image = gate.get_cpp_image(self.cpp_edep_image)

        # set the property of the output image:
        # in the coordinate system of the attached volume
        # FIXME no direction for the moment ?
        self.py_edep_image.SetOrigin(self.output_origin)

        # Uncertainty stuff need to be called before writing edep (to terminate temp events)
        if self.user_info.uncertainty:
            self.compute_uncertainty()
            n = gate.check_filename_type(self.user_info.output).replace(
                ".mhd", "_uncertainty.mhd"
            )
            itk.imwrite(self.uncertainty_image, n)

        # dose in gray
        if self.user_info.gray:
            self.py_dose_image = gate.get_cpp_image(self.cpp_dose_image)
            self.py_dose_image.SetOrigin(self.output_origin)
            n = gate.check_filename_type(self.user_info.output).replace(
                ".mhd", "_dose.mhd"
            )
            itk.imwrite(self.py_dose_image, n)

        # write the image at the end of the run
        # FIXME : maybe different for several runs
        if self.user_info.output:
            itk.imwrite(
                self.py_edep_image, gate.check_filename_type(self.user_info.output)
            )

    def compute_uncertainty(self):
        self.py_temp_image = gate.get_cpp_image(self.cpp_temp_image)
        self.py_square_image = gate.get_cpp_image(self.cpp_square_image)
        self.py_last_id_image = gate.get_cpp_image(self.cpp_last_id_image)

        self.py_temp_image.SetOrigin(self.output_origin)
        self.py_square_image.SetOrigin(self.output_origin)
        self.py_last_id_image.SetOrigin(self.output_origin)

        # complete edep with temp values
        edep = itk.array_view_from_image(self.py_edep_image)
        tmp = itk.array_view_from_image(self.py_temp_image)
        edep = edep + tmp
        self.py_edep_image = gate.itk_image_view_from_array(edep)
        self.py_edep_image.CopyInformation(self.py_temp_image)

        # complete square with temp values
        square = itk.array_view_from_image(self.py_square_image)
        square = square + tmp * tmp
        self.py_square_image = gate.itk_image_view_from_array(square)
        self.py_square_image.CopyInformation(self.py_temp_image)

        # uncertainty image
        self.uncertainty_image = gate.create_image_like(self.py_edep_image)
        unc = itk.array_view_from_image(self.uncertainty_image)
        N = unc.size
        # unc = np.sqrt(1 / (N - 1) * (square / N - np.power(edep / N, 2)))
        unc = 1 / (N - 1) * (square / N - np.power(edep / N, 2))
        unc = np.ma.masked_array(unc, unc < 0)
        unc = np.ma.sqrt(unc)
        unc = np.divide(unc, edep / N, out=np.ones_like(unc), where=edep != 0)
        self.uncertainty_image = gate.itk_image_view_from_array(unc)
        self.uncertainty_image.CopyInformation(self.py_temp_image)
        self.uncertainty_image.SetOrigin(self.output_origin)

        # debug
        """itk.imwrite(self.py_square_image, "square.mhd")
        itk.imwrite(self.py_temp_image, "temp.mhd")
        itk.imwrite(self.py_last_id_image, "lastid.mhd")
        itk.imwrite(self.uncertainty_image, "uncer.mhd")"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from test038_gan_phsp_spect_gan_helpers import *

paths = gate.get_default_test_paths(__file__, "gate_test038_gan_phsp_spect")
paths.output_ref = paths.output_ref / "test038_ref"

# create the simulation
sim = gate.Simulation()
condition_generator = create_simulation(sim, paths, None)

# add AA
gsource = sim.get_source_user_info("gaga")
gsource.skip_policy = "SkipEvents"
gsource.direction.acceptance_angle.volumes = ["spect1"]
gsource.direction.acceptance_angle.intersection_flag = True
gsource.direction.acceptance_angle.normal_flag = False
gsource.direction.acceptance_angle.skip_policy = "SkipEvents"
# gsource.batch_size = 2e4

# change output names
stat = sim.get_actor_user_info("Stats")
stat.output = paths.output / "test038_gan_aa_stats.txt"
proj = sim.get_actor_user_info("Projection_spect1_crystal")
proj.output = paths.output / "test038_gan_aa_proj.mhd"
singles = sim.get_actor_user_info("Singles_spect1_crystal")
singles.output = paths.output / "test038_gan_aa_singles.root"

# go (cannot be spawn in another process)
output = sim.start(False)

#
print()
s = output.get_source("gaga")
ref_se = 220534
t_se = (ref_se - s.fTotalSkippedEvents) / ref_se * 100
tol = 10
is_ok = t_se < tol
gate.print_test(
    is_ok,
    f"Source, nb of skipped particles (absorbed) : {s.fTotalSkippedEvents} {t_se:0.2f}% (tol = {tol}, {ref_se})",
)

print(
    f"Source, nb of zeros particles (absorbed) : {s.fTotalZeroEvents} (should be around 5)"
)

stats = output.get_actor("Stats")
stats_ref = gate.read_stat_file(paths.output_ref / "test038_gan_aa_stats.txt")
# do not compare steps
stats_ref.counts.step_count = stats.counts.step_count
is_ok = gate.assert_stats(stats, stats_ref, 0.02) and is_ok

stats.counts.event_count += s.fTotalSkippedEvents
print("Number of events is increased by the nb of skipped events")
print(stats)

# image
is_ok = (
    gate.assert_images(
        paths.output_ref / "test038_gan_aa_proj.mhd",
        paths.output / "test038_gan_aa_proj.mhd",
        stats,
        tolerance=70,
        axis="x",
        sum_tolerance=2,
    )
    and is_ok
)

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4
from scipy.spatial.transform import Rotation
import uproot
import matplotlib.pyplot as plt

paths = gate.get_default_test_paths(__file__, "gate_test036_adder_depth")


def create_simulation(geom):
    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.number_of_threads = 1
    ui.random_seed = 123456

    # units
    m = gate.g4_units("m")
    cm = gate.g4_units("cm")
    nm = gate.g4_units("nm")
    keV = gate.g4_units("keV")
    mm = gate.g4_units("mm")
    Bq = gate.g4_units("Bq")
    kBq = 1000 * Bq

    # world size
    world = sim.world
    world.size = [2 * m, 2 * m, 2 * m]

    # material
    sim.add_material_database(paths.data / "GateMaterials.db")

    # fake spect head
    head = sim.add_volume("Box", "SPECThead")
    head.size = [55 * cm, 42 * cm, 18 * cm]
    head.translation = [0, 0, 15 * cm]  ## not use if array of 2 heads
    head.material = "G4_AIR"

    # crystal
    crystal = sim.add_volume("Box", "crystal")
    crystal.mother = "SPECThead"
    crystal.size = [55 * cm, 42 * cm, 2 * cm]
    crystal.translation = [0, 0, 4 * cm]
    crystal.material = "Plastic"
    crystal.color = [1, 0, 0, 1]

    # pixel crystal
    crystal_pixel = sim.add_volume("Box", "crystal_pixel")
    crystal_pixel.mother = crystal.name
    crystal_pixel.size = [0.5 * cm, 0.5 * cm, 2 * cm]
    crystal_pixel.material = "NaITl"
    crystal_pixel.color = [1, 1, 0, 1]

    # geom
    size = [110, 84, 1]
    tr = [0.5 * cm, 0.5 * cm, 0]

    if geom == "repeat":
        le = gate.repeat_array(crystal_pixel.name, size, tr)
        crystal_pixel.translation = None
        crystal_pixel.rotation = None
        crystal_pixel.repeat = le

    if geom == "param":
        crystal_repeater = gate.build_param_repeater(
            sim, crystal.name, crystal_pixel.name, size, tr
        )

    # FIXME add a second head
    head.translation = None
    head.rotation = None
    tr = 30 * cm
    le = gate.repeat_array(head.name, [1, 1, 2], [0, 0, tr])
    le[0]["rotation"] = Rotation.from_euler("X", 180, degrees=True).as_matrix()
    head.repeat = le

    # physic list
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    p.enable_decay = False
    cuts = p.production_cuts
    cuts.world.gamma = 0.01 * mm
    cuts.world.electron = 0.01 * mm
    cuts.world.positron = 1 * mm
    cuts.world.proton = 1 * mm

    # default source for tests
    activity = 40 * kBq / ui.number_of_threads
    # activity = 5 * Bq / ui.number_of_threads
    source = sim.add_source("GenericSource", "src1")
    source.particle = "gamma"
    source.energy.mono = 333 * keV
    source.position.type = "sphere"
    source.position.radius = 5 * cm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, 1]
    source.activity = activity

    # default source for tests
    source = sim.add_source("GenericSource", "src2")
    source.particle = "gamma"
    source.energy.mono = 222 * keV
    source.position.type = "sphere"
    source.position.radius = 5 * cm
    source.direction.type = "momentum"
    source.direction.momentum = [0, 0, -1]
    source.activity = activity

    # add stat actor
    sim.add_actor("SimulationStatisticsActor", "Stats")

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    hc.mother = crystal.name
    hc.output = paths.output / "test036.root"
    hc.attributes = [
        "KineticEnergy",
        "PostPosition",
        # 'HitPosition', 'PrePosition',
        "TotalEnergyDeposit",
        "GlobalTime",  # 'EventID',
        # 'TrackVolumeName', 'TrackID',  # 'Test',
        # 'ProcessDefinedStep',
        "PreStepUniqueVolumeID",
        # 'TrackVolumeCopyNo', 'TrackVolumeInstanceID'
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.mother = crystal.name
    sc.input_digi_collection = "Hits"
    # sc.policy = 'EnergyWinnerPosition'
    sc.policy = "EnergyWeightedCentroidPosition"
    # same filename, there will be two branches in the file
    sc.output = hc.output

    sec = gate.g4_units("second")
    ui.running_verbose_level = 2
    # sim.run_timing_intervals = [[0, 0.33 * sec], [0.33 * sec, 0.66 * sec], [0.66 * sec, 1 * sec]]
    sim.run_timing_intervals = [[0, 1 * sec]]

    # print cuts
    pm = sim.physics_manager
    print(pm.dump_cuts())

    return sim


def test_output(output):
    # retrieve the information about the touched volumes
    man = g4.GateUniqueVolumeIDManager.GetInstance()
    vols = man.GetAllVolumeIDs()
    print(f"There are {len(vols)} volumes used in the adder")
    """for v in vols:
        vid = v.GetVolumeDepthID()
        print(f'Volume {v.fID}: ', end='')
        for x in vid:
            print(f' {x.fDepth} {x.fVolumeName} {x.fCopyNb} / ', end='')
        print()"""

    # stat
    gate.warning("Compare stats")
    stats = output.get_actor("Stats")
    print(stats)
    print(f"Number of runs was {stats.counts.run_count}. Set to 1 before comparison")
    stats.counts.run_count = 1  # force to 1
    stats_ref = gate.read_stat_file(paths.gate_output / "stats.txt")
    is_ok = gate.assert_stats(stats, stats_ref, tolerance=0.07)

    # root compare HITS
    print()
    hc = output.get_actor("Hits").user_info
    gate.warning("Compare HITS")
    gate_file = paths.gate_output / "spect.root"
    checked_keys = ["posX", "posY", "posZ", "edep", "time", "trackId"]
    keys1, keys2, scalings2, tols = gate.get_keys_correspondence(checked_keys)
    scalings = [1.0] * len(scalings2)
    tols[2] = 2  # Z
    # tols[4] = 0.01  # energy
    is_ok = gate.compare_root3(
        gate_file,
        hc.output,
        "Hits",
        "Hits",
        keys1,
        keys2,
        tols,
        scalings,
        scalings2,
        paths.output / "test036_hits.png",
    )

    # Root compare SINGLES
    print()
    sc = output.get_actor("Singles").user_info
    gate.warning("Compare SINGLES")
    gate_file = paths.gate_output / "spect.root"
    checked_keys = ["time", "globalPosX", "globalPosY", "globalPosZ", "energy"]
    keys1, keys2, scalings2, tols = gate.get_keys_correspondence(checked_keys)
    scalings = [1.0] * len(scalings2)
    tols[3] = 0.9  # Z
    # tols[1] = 1.0  # X
    # tols[2] = 1.0  # Y
    # tols[4] = 0.02  # energy
    is_ok = (
        gate.compare_root3(
            gate_file,
            sc.output,
            "Singles",
            "Singles",
            keys1,
            keys2,
            tols,
            scalings,
            scalings2,
            paths.output / "test036_singles.png",
        )
        and is_ok
    )

    # this is the end, my friend
    return is_ok



import opengate as gate
import opengate_core as g4
import uuid
from box import Box
from datetime import datetime


class SimulationStatisticsActor(g4.GateSimulationStatisticsActor, gate.ActorBase):
    """
    Store statistics about a simulation run.
    """

    type_name = "SimulationStatisticsActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.track_types_flag = False
        user_info.output = ""

    def __init__(self, user_info=None):
        # need to initialize (sometimes it is read from disk)
        self.simulation = None
        # user_info can be null when create empty actor (that read file)
        if not user_info:
            user_info = gate.UserInfo(
                "Actor", self.type_name, name=uuid.uuid4().__str__()
            )
        gate.ActorBase.__init__(self, user_info)
        g4.GateSimulationStatisticsActor.__init__(self, user_info.__dict__)
        actions = {"EndSimulationAction"}
        self.AddActions(actions)
        # actions are also set from the cpp side
        # empty results for the moment
        self.counts = Box()
        self.counts.run_count = 0
        self.counts.event_count = 0
        self.counts.track_count = 0
        self.counts.step_count = 0
        self.counts.duration = 0
        self.counts.start_time = 0
        self.counts.stop_time = 0
        self.counts.init = 0
        self.counts.track_types = {}

    def __del__(self):
        # print("del SimulationStatisticsActor", self.user_info.name)
        pass

    @property
    def pps(self):
        sec = gate.g4_units("s")
        if self.counts.duration != 0:
            return self.counts.event_count / self.counts.duration * sec
        return 0

    @property
    def tps(self):
        sec = gate.g4_units("s")
        if self.counts.duration != 0:
            return self.counts.track_count / self.counts.duration * sec
        return 0

    @property
    def sps(self):
        sec = gate.g4_units("s")
        if self.counts.duration != 0:
            return self.counts.step_count / self.counts.duration * sec
        return 0

    @property
    def nb_thread(self):
        if self.simulation is not None:
            thread = self.simulation.user_info.number_of_threads
        else:
            thread = "?"
        return thread

    @property
    def simu_start_time(self):
        if not self.simulation is None:
            sim_start = self.simulation.run_timing_intervals[0][0]
        else:
            sim_start = 0
        return sim_start

    @property
    def simu_end_time(self):
        if not self.simulation is None:
            sim_end = self.simulation.run_timing_intervals[-1][1]
        else:
            sim_end = 0
        return sim_end

    def __str__(self):
        if not self.counts:
            return ""
        sec = gate.g4_units("second")
        s = (
            f"Runs      {self.counts.run_count}\n"
            f"Events    {self.counts.event_count}\n"
            f"Tracks    {self.counts.track_count}\n"
            f"Step      {self.counts.step_count}\n"
            f'Init      {self.counts.init / sec} \t{g4.G4BestUnit(self.counts.init, "Time")}\n'
            f'Duration  {self.counts.duration / sec} \t{g4.G4BestUnit(self.counts.duration, "Time")}\n'
            f"PPS       {self.pps:.0f}\n"
            f"TPS       {self.tps:.0f}\n"
            f"SPS       {self.sps:.0f}\n"
            f"start     {self.counts.start_time}\n"
            f"stop      {self.counts.stop_time}\n"
            f'Sim start {g4.G4BestUnit(self.simu_start_time, "Time")}\n'
            f'Sim end   {g4.G4BestUnit(self.simu_end_time, "Time")}\n'
            f"Threads   {self.nb_thread}"
        )
        if self.user_info.track_types_flag:
            s += f"\n" f"Track types: {self.counts.track_types}"
        return s

    def EndSimulationAction(self):
        g4.GateSimulationStatisticsActor.EndSimulationAction(self)
        self.counts = Box(self.GetCounts())
        # write the file if an output filename was set
        if self.user_info.output != "":
            self.write(self.user_info.output)

    """
        It is feasible to get callback every Run, Event, Track, Step in the python side.
        However, it is VERY time consuming. For SteppingAction, expect large performance drop.
        It could be however useful for prototyping or tests.

        it requires "trampoline functions" on the cpp side.

        # feasible but very slow !
        def SteppingAction(self, step, touchable):
            g4.GateSimulationStatisticsActor.SteppingAction(self, step, touchable)
            do_something()
    """

    def write(self, filename):
        """
        Attempt to be mostly compatible to previous Gate stat output file
        """
        sec = gate.g4_units("s")
        f = open(filename, "w+")
        s = f"# NumberOfRun    = {self.counts.run_count}\n"
        s += f"# NumberOfEvents = {self.counts.event_count}\n"
        s += f"# NumberOfTracks = {self.counts.track_count}\n"
        s += f"# NumberOfSteps  = {self.counts.step_count}\n"
        s += f"# NumberOfGeometricalSteps  = ?\n"
        s += f"# NumberOfPhysicalSteps     = ?\n"
        s += f"# ElapsedTime           = {self.counts.duration / sec + self.counts.init / sec}\n"
        s += f"# ElapsedTimeWoInit     = {self.counts.duration / sec}\n"
        s += (
            f'# StartDate             = {g4.G4BestUnit(self.simu_start_time, "Time")}\n'
        )
        s += f'# EndDate               = {g4.G4BestUnit(self.simu_end_time, "Time")}\n'
        s += f"# PPS (Primary per sec)      = {self.pps:.0f}\n"
        s += f"# TPS (Track per sec)        = {self.tps:.0f}\n"
        s += f"# SPS (Step per sec)         = {self.sps:.0f}\n"
        s += f"# Threads                    = {self.nb_thread}\n"
        s += f"# Date                       = {datetime.now()}\n"
        if self.user_info.track_types_flag:
            s += f"# Track types:\n"
            for t in self.counts.track_types:
                s += f"# {t} = {self.counts.track_types[t]}\n"
        f.write(s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import itk

paths = gate.get_default_test_paths(__file__)
paths.data = paths.data / "t048_split_spect_projections"

# input of 4 heads projection images, with 3 energy windows
input_filenames = []
for i in range(4):
    input_filenames.append(paths.data / f"ref_projection_{i}.mhd")
print(f"Reading {len(input_filenames)} images")

info = gate.read_image_info(input_filenames[0])
print("Image info: ", info.size, info.spacing, info.origin)

# compute the output images
nb_ene = 3
outputs = gate.split_spect_projections(input_filenames, nb_ene)

# write them
e = 0
output_filenames = []
output_filename = str(paths.output / "t048_projection.mhd")
for o in outputs:
    f = output_filename.replace(".mhd", f"_{e}.mhd")
    itk.imwrite(o, f)
    output_filenames.append(f)
    e += 1

# ---------------------------------------------------------------
print()
gate.warning("Compare image to reference")
is_ok = True
for i in range(nb_ene):
    is_ok = (
        gate.assert_images(
            paths.output_ref / f"t048_projection_{i}.mhd",
            output_filenames[i],
            None,
            tolerance=1e-6,
            ignore_value=0,
            axis="x",
        )
        and is_ok
    )

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import opengate as gate
import itk
import numpy as np
from scipy.spatial.transform import Rotation


def main():
    paths = gate.get_default_test_paths(__file__, "gate_test042_gauss_gps")
    out_path = paths.output

    # create the simulation
    sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.g4_verbose_level = 1
    ui.visu = False
    ui.random_seed = 123456

    # units
    mm = gate.g4_units("mm")
    cm = gate.g4_units("cm")

    # add a material database
    sim.add_material_database(paths.gate_data / "HFMaterials2014.db")

    #  change world size
    world = sim.world
    world.size = [600 * cm, 500 * cm, 500 * cm]
    world.material = "Vacuum"

    geometries = ["Box", "Tubs"]
    grid_sizes = [[1, 1, 200], [1, 100, 200], [50, 100, 200]]
    res = 1 * mm
    spacing = [res, res, res]
    box_side = max(max(grid_sizes)) * res

    dose_actors = {}

    i = -10
    for geom in geometries:
        print(f"geom={geom}")
        for grid_size in grid_sizes:
            print(f"  grid_size={grid_size}")
            if geom == "Box":
                size = [s * res for s in grid_size]
            if geom == "Tubs":
                size = [
                    grid_size[0] * res,
                    0,
                    grid_size[2] * res / 2,
                ]  # [rmax, rmin, half_height]

            # mother not rotated, daughter not rotated
            print("    mother not rotated, daughter not rotated")
            transl = [i * box_side + 5 * mm, 0, 0]  # to avoid volume overlapping
            m = add_volume(
                sim, "Box", f"mother_{i}", [box_side, box_side, box_side], transl=transl
            )
            v = add_volume(
                sim, geom, f"vol_{i}", size, color=[1, 0, 1, 1], mother=m.name
            )
            d = add_dose_actor(sim, f"dose_{i}", grid_size, spacing, v.name, out_path)
            dose_actors[
                f"{geom}_{grid_size}_mother_not_rotated_daughter_not_rotated"
            ] = d
            i = i + 1

            # mother not rotated, daughter rotated
            print("    mother not rotated, daughter rotated")
            transl = [i * box_side + 5 * mm, 0, 0]  # to avoid volume overlapping
            m = add_volume(
                sim, "Box", f"mother_{i}", [box_side, box_side, box_side], transl=transl
            )
            rot = Rotation.from_euler("y", 90, degrees=True).as_matrix()
            v = add_volume(
                sim, geom, f"vol_{i}", size, color=[1, 0, 1, 1], rot=rot, mother=m.name
            )
            d = add_dose_actor(sim, f"dose_{i}", grid_size, spacing, v.name, out_path)
            dose_actors[f"{geom}_{grid_size}_mother_not_rotated_daughter_rotated"] = d
            i = i + 1

            # mother rotated, daughter not rotated
            print("    mother rotated, daughter not rotated")
            rot = Rotation.from_euler("y", 90, degrees=True).as_matrix()
            transl = [i * box_side + 5 * mm, 0, 0]  # to avoid volume overlapping
            m = add_volume(
                sim,
                "Box",
                f"mother_{i}",
                [box_side, box_side, box_side],
                rot=rot,
                transl=transl,
            )
            v = add_volume(
                sim, geom, f"vol_{i}", size, color=[1, 0, 1, 1], mother=m.name
            )
            d = add_dose_actor(sim, f"dose_{i}", grid_size, spacing, v.name, out_path)
            dose_actors[f"{geom}_{grid_size}_mother_rotated_daughter_not_rotated"] = d
            i = i + 1

            # mother rotated, daughter rotated
            print("    mother rotated, daughter rotated")
            rot = Rotation.from_euler("y", 90, degrees=True).as_matrix()
            transl = [i * box_side + 5 * mm, 0, 0]  # to avoid volume overlapping
            m = add_volume(
                sim,
                "Box",
                f"mother_{i}",
                [box_side, box_side, box_side],
                rot=rot,
                transl=transl,
            )
            rot = Rotation.from_euler("x", 90, degrees=True).as_matrix()
            v = add_volume(
                sim, geom, f"vol_{i}", size, color=[1, 0, 1, 1], rot=rot, mother=m.name
            )
            d = add_dose_actor(sim, f"dose_{i}", grid_size, spacing, v.name, out_path)
            dose_actors[f"{geom}_{grid_size}_mother_rotated_daughter_rotated"] = d
            i = i + 1

    # start simulation
    output = sim.start()

    # test
    ok = True
    for test_name, dose in dose_actors.items():
        print(test_name)
        ok = check_dose_grid_geometry(dose.output, dose) and ok

    gate.test_ok(ok)


def check_dose_grid_geometry(dose_mhd_path, dose_actor):
    img = itk.imread(dose_mhd_path)
    data = itk.GetArrayViewFromImage(img)
    shape = data.shape
    spacing = img.GetSpacing()
    shape_ref = tuple(np.flip(dose_actor.size))
    spacing_ref = dose_actor.spacing

    ok = True
    if shape != shape_ref:
        print(f"    shape={shape} not the same as shape_ref{shape_ref}!")
        ok = False
    else:
        print("    Shape ok")

    if spacing != spacing_ref:
        print(f"    spacing={spacing} not the same as spacing_ref={spacing_ref}!")
        ok = False
    else:
        print("    Spacing ok")

    return ok


def add_volume(
    sim, geom, name, size, color=[0, 0, 1, 1], rot=None, transl=[0, 0, 0], mother=None
):
    phantom = sim.add_volume(geom, name)
    if mother:
        phantom.mother = mother
    if geom == "Box":
        phantom.size = size
    if geom == "Tubs":
        phantom.rmax = size[0]
        phantom.rmin = size[1]
        phantom.dz = size[2]
    if rot is not None:
        phantom.rotation = rot
    phantom.translation = transl
    phantom.material = "G4_WATER"
    phantom.color = color

    return phantom


def add_dose_actor(sim, name, size, spacing, mother, out_path):
    dose = sim.add_actor("DoseActor", name)
    dose.output = out_path / f"{name}.mhd"
    dose.mother = mother
    dose.size = size
    dose.spacing = spacing
    dose.hit_type = "random"

    return dose


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate

paths = gate.get_default_test_paths(__file__, "gate_test008_dose_actor")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.random_seed = 123456

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")

#  change world size
world = sim.world
world.size = [0.5 * m, 0.5 * m, 0.5 * m]

# waterbox
waterbox = sim.add_volume("Box", "waterbox")
waterbox.size = [10 * cm, 10 * cm, 10 * cm]
waterbox.material = "G4_WATER"
waterbox.color = [0, 0, 1, 1]

# lungbox
lungbox = sim.add_volume("Box", "lungbox")
lungbox.mother = waterbox.name
lungbox.size = [10 * cm, 10 * cm, 4 * cm]
lungbox.translation = [0 * cm, 0 * cm, 2.5 * cm]
lungbox.material = "G4_LUNG_ICRP"
lungbox.color = [0, 1, 1, 1]

# bonebox
bonebox = sim.add_volume("Box", "bonebox")
bonebox.mother = waterbox.name
bonebox.size = [10 * cm, 10 * cm, 4 * cm]
bonebox.translation = [0 * cm, 0 * cm, -2.5 * cm]
bonebox.material = "G4_BONE_CORTICAL_ICRP"
bonebox.color = [1, 0, 0, 1]

# physics
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMV"
sim.set_cut("world", "all", 1 * mm)

# default source for tests
source = sim.add_source("GenericSource", "mysource")
source.energy.mono = 115 * MeV
source.particle = "proton"
source.position.type = "disc"
source.position.radius = 1 * cm
source.position.translation = [0, 0, -80 * mm]
source.direction.type = "momentum"
source.direction.momentum = [0, 0, 1]
source.activity = 5000 * Bq

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test041-edep.mhd"
dose.mother = "waterbox"
dose.size = [10, 10, 50]
mm = gate.g4_units("mm")
ts = [200 * mm, 200 * mm, 200 * mm]
dose.spacing = [x / y for x, y in zip(ts, dose.size)]
print(dose.spacing)
dose.uncertainty = True
dose.gray = True
dose.hit_type = "random"

# add stat actor
s = sim.add_actor("SimulationStatisticsActor", "Stats")
s.track_types_flag = True

# start simulation
output = sim.start(True)

# print results at the end
stat = output.get_actor("Stats")
print(stat)

dose = output.get_actor("dose")
print(dose)

# tests
gate.warning("Tests stats file")
stats_ref = gate.read_stat_file(paths.gate_output / "stat2.txt")
is_ok = gate.assert_stats(stat, stats_ref, 0.10)

gate.warning("\nDifference for EDEP")
is_ok = (
    gate.assert_images(
        paths.gate_output / "output2-Edep.mhd",
        paths.output / "test041-edep.mhd",
        stat,
        tolerance=10,
        ignore_value=0,
    )
    and is_ok
)

gate.warning("\nDifference for uncertainty")
is_ok = (
    gate.assert_images(
        paths.gate_output / "output2-Edep-Uncertainty.mhd",
        paths.output / "test041-edep_uncertainty.mhd",
        stat,
        tolerance=30,
        ignore_value=1,
    )
    and is_ok
)

gate.warning("\nDifference for dose in Gray")
is_ok = (
    gate.assert_images(
        paths.gate_output / "output2-Dose.mhd",
        paths.output / "test041-edep_dose.mhd",
        stat,
        tolerance=10,
        ignore_value=0,
    )
    and is_ok
)

gate.test_ok(is_ok)



import opengate as gate
import os
import numpy as np
import opengate_core as g4


def read_voxel_materials(filename, def_mat="G4_AIR"):
    p = os.path.abspath(filename)
    f = open(p, "r")
    current = 0
    materials = []
    for line in f:
        for word in line.split():
            if word[0] == "#":
                break
            if current == 0:
                start = float(word)
                current = 1
            else:
                if current == 1:
                    stop = float(word)
                    current = 2
                else:
                    if current == 2:
                        mat = word
                        current = 0
                        materials.append([start, stop, mat])

    # sort according to starting interval
    materials = sorted(materials)

    # consider all values
    pix_mat = []
    previous = None
    for m in materials:
        if previous and previous > m[0]:
            gate.fatal(
                f"Error while reading {filename}\n"
                f"Intervals are not disjoint: {previous} {m}"
            )
        if m[0] > m[1]:
            gate.fatal(f"Error while reading {filename}\n" f"Wrong interval {m}")
        if not previous or previous == m[0]:
            pix_mat.append([previous, m[1], m[2]])
            previous = m[1]
        else:
            pix_mat.append([previous, m[0], def_mat])
            pix_mat.append([previous, m[1], m[2]])
            previous = m[1]

    return pix_mat


def new_material_weights(name, density, elements, weights=[1]):
    n = g4.G4NistManager.Instance()
    if not isinstance(elements, list):
        elements = [elements]
    if len(elements) != len(weights):
        gate.fatal(
            f"Cannot create the new material, the elements and the "
            f"weights does not have the same size: {elements} and {weights}"
        )
    total = np.sum(weights)
    weights = weights / total
    m = n.ConstructNewMaterialWeights(name, elements, weights, density)
    return m


def new_material_nb_atoms(name, density, elements, nb_atoms):
    n = g4.G4NistManager.Instance()
    if not isinstance(elements, list):
        elements = [elements]
    if len(elements) != len(nb_atoms):
        gate.fatal(
            f"Cannot create the new material, the elements and the "
            f"nb_atoms does not have the same size: {elements} and {nb_atoms}"
        )
    nb_atoms = [int(x) for x in nb_atoms]
    m = n.ConstructNewMaterialNbAtoms(name, elements, nb_atoms, density)
    return m


def HU_read_materials_table(file_mat):
    p = os.path.abspath(file_mat)
    f = open(p, "r")
    elements = ["HU"]
    materials = []
    current_section = None
    current_material = None
    for line in f:
        i = 0
        for word in line.split():
            if word[0] == "#":
                break
            if word == "[Elements]":
                current_section = "element"
                break
            if word == "[/Elements]":
                elements.append("name")
                current_section = "table"
                break
            if current_section is None:
                break
            if current_section == "element":
                elements.append(word)
            if current_section == "table":
                if current_material is None:
                    current_material = {}
                if i == 0:
                    current_material[elements[i]] = int(word)
                else:
                    if i == len(elements) - 1:
                        current_material[elements[i]] = word
                    else:
                        current_material[elements[i]] = float(word)
            i += 1
        if current_material:
            materials.append(current_material)
        current_material = None
    return materials, elements


def HU_read_density_table(file_density):
    p = os.path.abspath(file_density)
    f = open(p, "r")
    densities = []
    for line in f:
        words = line.split()
        if len(words) < 1:
            continue
        if words[0][0] == "#":
            continue
        d = {"HU": int(words[0]), "density": float(words[1])}
        densities.append(d)
    return densities


def HU_linear_interpolate_densities(hu, densities):
    i = 0
    n = len(densities)
    while i < n and hu > densities[i]["HU"]:
        i = i + 1
    i = i - 1
    if i < 0:
        return densities[0]["density"]
    if i >= n - 1:
        return densities[n - 1]["density"]
    v = ((hu - densities[i]["HU"]) / (densities[i + 1]["HU"] - densities[i]["HU"])) * (
        densities[i + 1]["density"] - densities[i]["density"]
    ) + densities[i]["density"]
    return v


def HU_find_max_density_difference(hu_min, hu_max, d_min, d_max, densities):
    n = len(densities)
    i = 0
    while i < n and hu_min > densities[i]["HU"]:
        i = i + 1
    j = 0
    while j < n and hu_max > densities[j]["HU"]:
        j = j + 1
    j = j - 1
    for x in range(i, j, 1):
        if densities[i]["density"] < d_min:
            d_min = densities[i]["density"]
        if densities[i]["density"] > d_max:
            d_max = densities[i]["density"]
    return d_max - d_min


# correspondence element names <> symbol
elements_name_symbol = {
    "Hydrogen": "H",
    "Carbon": "C",
    "Nitrogen": "N",
    "Oxygen": "O",
    "Sodium": "Na",
    "Magnesium": "Mg",
    "Phosphor": "P",
    "Sulfur": "S",
    "Chlorine": "Cl",
    "Argon": "Ar",
    "Potassium": "K",
    "Calcium": "Ca",
    "Titanium": "Ti",
    "Copper": "Cu",
    "Zinc": "Zn",
    "Silver": "Ag",
    "Tin": "Sn",
}


def HounsfieldUnit_to_material(density_tolerance, file_mat, file_density):
    """
    Same function than in GateHounsfieldToMaterialsBuilder class.
    Probably far from optimal, put we keep the compatibility
    """

    materials, elements = HU_read_materials_table(file_mat)
    densities = HU_read_density_table(file_density)
    voxel_materials = []
    created_materials = []
    gcm3 = gate.g4_units("g/cm3")

    elems = elements[1 : len(elements) - 1]
    elems_symbol = [elements_name_symbol[x] for x in elems]

    i = 0
    num = 0
    last_i = len(materials) - 1
    nm = g4.G4NistManager.Instance()
    for mat in materials:
        # get HU interval
        hu_min = mat["HU"]
        if i == last_i:
            hu_max = hu_min + 1
        else:
            hu_max = materials[i + 1]["HU"]

        # check hu min max
        if hu_max <= hu_min:
            gate.fatal(f"Error, HU interval not valid: {mat}")

        # get densities interval
        dmin = HU_linear_interpolate_densities(hu_min, densities)
        dmax = HU_linear_interpolate_densities(hu_max, densities)
        ddiff = HU_find_max_density_difference(hu_min, hu_max, dmin, dmax, densities)

        # nb of bins
        n = max(1, ddiff * gcm3 / density_tolerance)
        # n_naive = max(1, (dmax - dmin) * gcm3 / density_tolerance)

        # check if AIR
        if "Air" in mat["name"] or "AIR" in mat["name"]:
            n = 1

        # HU interval according to tolerance
        htol = (hu_max - hu_min) / n

        # like in Gate
        if n != 1:
            n = int(np.ceil(n))

        # loop on density interval
        for j in range(n):
            h1 = hu_min + j * htol
            h2 = min(hu_min + (j + 1) * htol, hu_max)
            d = HU_linear_interpolate_densities(h1 + (h2 - h1) / 2.0, densities)
            # create a new material with the interpolated density
            weights = [mat[x] for x in elems]
            weights_nz = []
            elems_symbol_nz = []
            # remove the weight equal to zero
            sum = 0
            for a, e in zip(weights, elems_symbol):
                if a > 0:
                    weights_nz.append(a)
                    elems_symbol_nz.append(e)
                    sum += a
            # normalise weight
            for k in range(len(weights_nz)):
                weights_nz[k] = weights_nz[k] / sum
            # create a new material
            m = nm.ConstructNewMaterialWeights(
                f'{mat["name"]}_{num}', elems_symbol_nz, weights_nz, d * gcm3
            )
            # get the final correspondence
            c = [h1, h2, str(m.GetName())]
            voxel_materials.append(c)
            created_materials.append(m)
            num = num + 1
        #
        i = i + 1
    return voxel_materials, created_materials


def dump_material_like_Gate(mat):
    s = f'{mat.GetName()}: d={gate.g4_best_unit(mat.GetDensity(), "Volumic Mass")}; n={mat.GetNumberOfElements()}\n'
    i = 0
    for elem in mat.GetElementVector():
        s += f"+el: name={elem.GetName()}; f={mat.GetElementFraction(i)}\n"
        i += 1
    s += "\n"
    return s


def assert_same_material(m1, m2):
    if m1.name != m2.name:
        return False
    if np.fabs(m1.density - m2.density) / m1.density > 1e-2:
        print("Error while comparing materials", m1, m2)
        print(np.fabs(m1.density - m2.density) / m1.density)
        print(m1)
        print(m2)
        return False
    for e1 in m1.components:
        e2 = m2.components[elements_name_symbol[e1]]
        e1 = m1.components[e1]
        if elements_name_symbol[e1.name] != e2.name:
            print("Error while comparing materials", m1, m2)
            print(e1, e2)
            return False
        if e1.n != e2.n:
            print("Error while comparing materials", m1, m2)
            print(e1, e2)
            return False
        if np.fabs(e1.f - e2.f) / e1.f > 1e-2:
            print("Error while comparing materials", m1, m2)
            print(e1, e2)
            return False

    return True


def read_next_line(f):
    line = f.readline()
    return line.strip().replace("\t", " ")


def read_tag(s, tag):
    w = s.split("=")
    if w[0].strip() != tag:
        return None
    value = w[1].strip()
    return value


def read_tag_with_unit(s, tag):
    w = s.split("=")
    if w[0].strip() != tag:
        return None
    w = w[1].split()
    value = float(w[0])
    u = gate.g4_units(w[1].strip())
    return value * u



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate_core as g4
import math

# Get Nist Material Manager
n = g4.G4NistManager.Instance()
print("Nist manager", n)

# http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Appendix/materialNames.html
mat = n.FindOrBuildMaterial("G4_WATER")
print("Mat Water", mat)
assert mat.GetName() == "G4_WATER"
gcm3 = gate.g4_units("g/cm3")
print("Density ", mat.GetDensity(), mat.GetDensity() / gcm3)
assert math.isclose(mat.GetDensity(), gcm3)
print("Elements", mat.GetElementVector())
print("Nb of elements", mat.GetNumberOfElements())
assert mat.GetNumberOfElements() == 2
elements = mat.GetElementVector()
assert elements[0].GetSymbol() == "H"
assert elements[1].GetSymbol() == "O"

eV = gate.g4_units("eV")
Imean = mat.GetIonisation().GetMeanExcitationEnergy()
print("I mean = ", Imean / eV, "eV")
assert math.isclose(Imean / eV, 78.0)

# Another material
mat = n.FindOrBuildMaterial("G4_TISSUE-PROPANE")
print("Mat 2", mat, mat.GetName())
N = mat.GetElementVector()[2]
gmol = gate.g4_units("g/mol")
print("N Z", N.GetZ())
print("N A", N.GetA() / gmol)
assert N.GetZ() == 7
assert math.isclose(N.GetA() / gmol, 14.00676896)

# n.ListMaterials('all')

# simple simulation object
print("-" * 80)
sim = gate.Simulation()
print(sim)

gate.test_ok(True)



import opengate as gate
import pathlib

# colors
red = [1, 0, 0, 1]
blue = [0, 0, 1, 1]
green = [0, 1, 0, 1]
yellow = [0.9, 0.9, 0.3, 1]
gray = [0.5, 0.5, 0.5, 1]
white = [1, 1, 1, 0.8]
transparent = [1, 1, 1, 0]


def add_pet(sim, name="pet", load_db=True):
    """
    Geometry of a PET Siemens Biograph
    https://doi.org/10.1002/mp.16032
    """

    # unit
    mm = gate.g4_units("mm")

    # material
    if load_db:
        f = pathlib.Path(__file__).parent.resolve()
        sim.add_material_database(f / "pet_siemens_biograph_materials.db")

    # ring volume
    pet = sim.add_volume("Tubs", name)
    pet.rmax = 460 * mm
    pet.rmin = 410 * mm
    pet.dz = 270 * mm / 2.0
    pet.color = white
    pet.material = "G4_AIR"

    # sim.user_info.check_volumes_overlap = False
    sim.user_info.check_volumes_overlap = True

    # 4 rings
    ring = sim.add_volume("Tubs", f"{name}_ring")
    ring.mother = pet.name
    ring.rmax = 460 * mm
    ring.rmin = 410 * mm
    ring.dz = 56 * mm / 2
    ring.translation = None
    ring.rotation = None
    ring.material = "G4_AIR"
    le = gate.repeat_array(ring.name, [1, 1, 4], [0, 0 * mm, 56 * mm])
    ring.repeat = le
    ring.color = transparent

    # Block
    block = sim.add_volume("Box", f"{name}_block")
    block.mother = ring.name
    block.size = [20 * mm, 52 * mm, 52 * mm]
    block.translation = None
    block.rotation = None
    block.material = "VM2000"
    le = gate.repeat_ring(block.name, -4.28572, 48, [438 * mm, 0, 0], [0, 0, 1])
    block.color = blue
    block.repeat = le

    # Crystal
    crystal = sim.add_volume("Box", f"{name}_crystal")
    crystal.mother = block.name
    crystal.size = [20 * mm, 3.98 * mm, 3.98 * mm]
    crystal.material = "LSO"
    crystal.translation = None
    crystal.rotation = None
    le = gate.repeat_array(crystal.name, [1, 13, 13], [0, 4 * mm, 4 * mm])
    crystal.repeat = le
    crystal.color = red

    return pet


def add_digitizer(
    sim, pet_name, output_filename, hits_name="Hits", singles_name="Singles"
):
    # get crystal volume
    crystal = sim.get_volume_user_info(f"{pet_name}_crystal")
    block = sim.get_volume_user_info(f"{pet_name}_block")
    ring = sim.get_volume_user_info(f"{pet_name}_ring")

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", hits_name)
    hc.mother = crystal.name
    hc.output = output_filename
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "PreStepUniqueVolumeID",
        "GlobalTime",
    ]

    # singles collection
    sc = sim.add_actor("DigitizerReadoutActor", singles_name)
    sc.input_digi_collection = hc.name
    sc.group_volume = block.name
    # sc.group_volume = ring.name # (I checked that is it different from block)
    # sc.group_volume = crystal.name # (I checked that is it different from block)
    sc.discretize_volume = crystal.name
    sc.policy = "EnergyWeightedCentroidPosition"
    sc.output = hc.output

    return sc



from .BoxVolume import *
from .SphereVolume import *
from .TrapVolume import *
from .ImageVolume import *
from .TubsVolume import *
from .ConsVolume import *
from .PolyhedraVolume import *
from .HexagonVolume import *
from .TrdVolume import *
from .BooleanVolume import *
from .RepeatParametrisedVolume import *
from anytree import RenderTree
from anytree import Node
import copy

volume_type_names = {
    BoxVolume,
    SphereVolume,
    TrapVolume,
    ImageVolume,
    TubsVolume,
    PolyhedraVolume,
    HexagonVolume,
    ConsVolume,
    TrdVolume,
    BooleanVolume,
    RepeatParametrisedVolume,
}
volume_builders = gate.make_builders(volume_type_names)

# G4Tubs G4CutTubs G4Cons G4Para G4Trd
# G4Torus (G4Orb not needed) G4Tet
# G4EllipticalTube G4Ellipsoid G4EllipticalCone
# G4Paraboloid G4Hype
# specific: G4Polycone G4GenericPolycone Polyhedra
# G4ExtrudedSolid G4TwistedBox G4TwistedTrap G4TwistedTrd G4GenericTrap
# G4TwistedTubs

"""
http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html#constructed-solid-geometry-csg-solids
"""


def box_add_size(box, thickness):
    box.size = [x + thickness for x in box.size]


def cons_add_size(cons, thickness):
    cons.rmax1 += thickness / 2
    cons.rmax2 += thickness / 2
    cons.dz += thickness


def get_volume_bounding_limits(simulation, volume_name):
    """
    Return the min and max 3D points of the bounding box of the given volume
    """
    v = simulation.get_volume_user_info(volume_name)
    s = simulation.get_solid_info(v)
    pMin = s.bounding_limits[0]
    pMax = s.bounding_limits[1]
    return pMin, pMax


def get_volume_bounding_box_size(simulation, volume_name):
    """
    Return the size of the bounding box of the given volume
    """
    pMin, pMax = get_volume_bounding_limits(simulation, volume_name)
    return [pMax[0] - pMin[0], pMax[1] - pMin[1], pMax[2] - pMin[2]]


def translate_point_to_volume(simulation, volume, top, x):
    """

    Consider the point x in the current volume and return the coordinate of x in the top volume
    (that must be an ancestor).
    Translation only, do not consider rotation.
    """
    while volume.name != top:
        x += volume.translation
        volume = simulation.get_volume_user_info(volume.mother)
    return x


# correspondence element names <> symbol
elements_name_symbol = {
    "Hydrogen": "H",
    "Carbon": "C",
    "Nitrogen": "N",
    "Oxygen": "O",
    "Sodium": "Na",
    "Magnesium": "Mg",
    "Phosphor": "P",
    "Sulfur": "S",
    "Chlorine": "Cl",
    "Argon": "Ar",
    "Potassium": "K",
    "Calcium": "Ca",
    "Titanium": "Ti",
    "Copper": "Cu",
    "Zinc": "Zn",
    "Silver": "Ag",
    "Tin": "Sn",
}


def render_tree(tree, geometry):
    """
    Print a tree of volume
    """
    s = ""
    for pre, fill, node in RenderTree(tree[gate.__world_name__]):
        v = geometry[node.name]
        s += f"{pre}{node.name} {v.type_name} {v.material}\n"

    # remove last break line
    return s[:-1]


def build_tree(simulation):
    uiv = simulation.volume_manager.user_info_volumes
    # world is needed as the root
    if gate.__world_name__ not in uiv:
        s = f"No world in geometry = {uiv}"
        gate.fatal(s)

    # build the root tree (needed)
    tree = {gate.__world_name__: Node(gate.__world_name__)}
    already_done = {gate.__world_name__: True}

    # build the tree
    for vol in uiv.values():
        if vol.name in already_done:
            continue
        add_volume_to_tree(uiv, already_done, tree, vol)

    return tree


def add_volume_to_tree(user_info_volumes, already_done, tree, vol):
    # check if mother volume exists
    uiv = user_info_volumes
    if vol.mother not in uiv:
        gate.fatal(
            f"Cannot find a mother volume named '{vol.mother}', for the volume {vol}"
        )

    already_done[vol.name] = "in_progress"
    m = uiv[vol.mother]

    # check for the cycle
    if m.name not in already_done:
        add_volume_to_tree(uiv, already_done, tree, m)
    else:
        if already_done[m.name] == "in_progress":
            s = f"Error while building the tree, there is a cycle ? "
            s += f"\n volume is {vol}"
            s += f"\n parent is {m}"
            gate.fatal(s)

    # get the mother branch
    p = tree[m.name]

    # check not already exist
    if vol.name in tree:
        s = f"Node already exist in tree {vol.name} -> {tree}"
        s = s + f"\n Probably two volumes with the same name ?"
        gate.fatal(s)

    # create the node
    n = Node(vol.name, parent=p)
    tree[vol.name] = n
    already_done[vol.name] = True


def copy_volume_user_info(ref_volume, target_volume):
    for att in ref_volume.__dict__:
        if att != "_name":
            target_volume.__dict__[att] = copy.deepcopy(ref_volume.__dict__[att])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate_core as g4
import opengate as gate

# get the attributes manager
am = g4.GateDigiAttributeManager.GetInstance()

# retrieve the list of attributes names
nlist = am.GetAvailableDigiAttributeNames()

# print
print(f"List of all available attributes ({len(nlist)})")
print(
    f"Types are: 3 (ThreeVector), D (double), S (string), I (int), U (unique volume ID)"
)
for a in nlist:
    att = am.GetDigiAttributeByName(a)
    print(att.GetDigiAttributeName(), att.GetDigiAttributeType())

n = 38
is_ok = len(nlist) == n

gate.print_test(is_ok, f"Done for {n} attributes.")

gate.test_ok(is_ok)



import opengate as gate
from .VoxelizedSourcePDFSampler import *


class VoxelizedSourceConditionGenerator:
    def __init__(self, activity_source_filename, rs=np.random):
        self.activity_source_filename = str(activity_source_filename)
        self.image = None
        self.cdf_x = self.cdf_y = self.cdf_z = None
        self.rs = rs
        self.img_info = None
        self.sampler = None
        self.initialize_source()
        self.compute_directions = False

    def initialize_source(self):
        self.image = itk.imread(self.activity_source_filename)
        self.img_info = gate.get_info_from_image(self.image)
        self.sampler = VoxelizedSourcePDFSampler(self.image)
        self.rs = np.random

    def generate_condition(self, n):
        # i j k is in np array order = z y x
        # but img_info is in the order x y z
        i, j, k = self.sampler.sample_indices(n, self.rs)

        # half pixel size
        hs = self.img_info.spacing / 2.0

        # sample within the voxel
        rx = self.rs.uniform(-hs[0], hs[0], size=n)
        ry = self.rs.uniform(-hs[1], hs[1], size=n)
        rz = self.rs.uniform(-hs[2], hs[2], size=n)

        # warning order np is z,y,x while itk is x,y,z
        x = self.img_info.spacing[2] * k + rz
        y = self.img_info.spacing[1] * j + ry
        z = self.img_info.spacing[0] * i + rx

        # x,y,z are in the image coord system
        # we set in the g4 coord system: according to the center of the image
        p = np.column_stack((x, y, z)) - hs * self.img_info.size + hs

        # need direction ?
        if self.compute_directions:
            v = gate.generate_isotropic_directions(n, rs=self.rs)
            return np.column_stack((p, v))
        else:
            return p



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import itk
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__, "")

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.visu = False
ui.number_of_threads = 1
ui.random_seed = 123456
print(ui)

# add a material database
sim.add_material_database(paths.data / "GateMaterials.db")

# units
m = gate.g4_units("m")
mm = gate.g4_units("mm")
cm = gate.g4_units("cm")
keV = gate.g4_units("keV")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
kBq = 1000 * Bq

#  change world size
world = sim.world
world.size = [1.5 * m, 1 * m, 1 * m]

# fake box #1
fake = sim.add_volume("Box", "fake")
fake.size = [36 * cm, 36 * cm, 36 * cm]
fake.translation = [25 * cm, 0, 0]
r = Rotation.from_euler("y", -25, degrees=True)
r = r * Rotation.from_euler("x", -35, degrees=True)
fake.rotation = r.as_matrix()

# ---------------------------------------------------
# CT image #1
ct = sim.add_volume("Image", "ct")
ct.image = str(paths.data / "10x10x10.mhd")
ct.mother = fake.name
ct.voxel_materials = [[0, 10, "G4_WATER"]]
ct.translation = [-3 * cm, 0, 0]
r = Rotation.from_euler("z", 45, degrees=True)
ct.rotation = r.as_matrix()

ct_info = gate.read_image_info(ct.image)
print(f"CT image origin and size: ", ct_info.origin, ct_info.size, ct_info.spacing)

# source from image for CT #1
source = sim.add_source("VoxelsSource", "vox_source")
source.mother = ct.name
source.particle = "alpha"
source.activity = 10000 * Bq / ui.number_of_threads
source.image = str(paths.data / "five_pixels_10.mhd")
source.direction.type = "iso"
source.position.translation = gate.get_translation_between_images_center(
    ct.image, source.image
)
print(f"Source wrt CT 10x10x10 translation", source.position.translation)
source.energy.mono = 1 * MeV
src_info = gate.read_image_info(source.image)
print(
    f"Source image origin and size: ", src_info.origin, src_info.size, src_info.spacing
)

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test021-edep_1.mhd"
dose.mother = ct.name
img_info = gate.read_image_info(ct.image)
dose.size = img_info.size
dose.spacing = img_info.spacing
dose.img_coord_system = True

# cuts
p = sim.get_physics_user_info()
p.physics_list_name = "QGSP_BERT_EMZ"
p.enable_decay = False
sim.set_cut("world", "all", 1 * mm)

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# verbose
sim.apply_g4_command("/tracking/verbose 0")

# start simulation
output = sim.start()

# print results at the end
stat = output.get_actor("Stats")
# stat.write(paths.output_ref / "stat021_ref_1.txt")

# test pixels in dose #1
# test pixels in dose #1
d_even = itk.imread(str(dose.output))
s = itk.array_view_from_image(d_even).sum()
v0 = d_even.GetPixel([5, 5, 5])
v1 = d_even.GetPixel([1, 5, 5])
v2 = d_even.GetPixel([1, 2, 5])
v3 = d_even.GetPixel([5, 2, 5])
v4 = d_even.GetPixel([6, 2, 5])
tol = 0.15
ss = v0 + v1 + v2 + v3 + v4


def t(s, v):
    diff = abs(s - v) / s
    b = diff < tol
    p = diff * 100.0
    gate.print_test(b, f"Image diff {s:.2f} vs {v:.2f}  -> {p:.2f}%")
    return b


is_ok = t(s, ss)
is_ok = t(2000, v0) and is_ok
is_ok = t(2000, v1) and is_ok
is_ok = t(2000, v2) and is_ok
is_ok = t(2000, v3) and is_ok
is_ok = t(2000, v4) and is_ok

stats_ref = gate.read_stat_file(paths.output_ref / "stat021_ref_1.txt")
stats_ref.counts.run_count = ui.number_of_threads
is_ok = gate.assert_stats(stat, stats_ref, 0.1) and is_ok

gate.test_ok(is_ok)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
import opengate.contrib.pet_philips_vereos as pet_vereos
import opengate.contrib.phantom_necr as phantom_necr


def make_simu(sim=None, output_path="./"):
    # create the simulation
    if sim is None:
        sim = gate.Simulation()

    # main options
    ui = sim.user_info
    ui.g4_verbose = False
    ui.visu = False
    ui.check_volumes_overlap = False

    # units
    m = gate.g4_units("m")
    mm = gate.g4_units("mm")
    cm = gate.g4_units("cm")
    Bq = gate.g4_units("Bq")
    MBq = Bq * 1e6
    sec = gate.g4_units("second")

    #  change world size
    world = sim.world
    world.size = [3 * m, 3 * m, 3 * m]
    world.material = "G4_AIR"

    # add a PET VEREOS
    pet = pet_vereos.add_pet(sim, "pet")

    # add table
    bed = pet_vereos.add_table(sim, "pet")

    # add NECR phantom
    phantom = phantom_necr.add_necr_phantom(sim, "phantom")

    # physics
    p = sim.get_physics_user_info()
    p.physics_list_name = "G4EmStandardPhysics_option4"
    sim.set_cut("world", "all", 1 * m)
    sim.set_cut(phantom.name, "all", 0.1 * mm)
    sim.set_cut(bed.name, "all", 0.1 * mm)
    sim.set_cut(f"{pet.name}_crystal", "all", 0.1 * mm)

    # default source for tests
    source = phantom_necr.add_necr_source(sim, phantom)
    total_yield = gate.get_rad_yield("F18")
    print("Yield for F18 (nb of e+ per decay) : ", total_yield)
    source.activity = 3000 * Bq * total_yield
    source.activity = 1787.914158 * MBq * total_yield
    source.half_life = 6586.26 * sec
    source.energy.type = "F18_analytic"  # FIXME not ok, but similar to previous Gate

    # add stat actor
    s = sim.add_actor("SimulationStatisticsActor", "Stats")
    s.track_types_flag = True

    # hits collection
    hc = sim.add_actor("DigitizerHitsCollectionActor", "Hits")
    # get crystal volume by looking for the word crystal in the name
    l = sim.get_all_volumes_user_info()
    crystal = l[[k for k in l if "crystal" in k][0]]
    hc.mother = crystal.name
    print("Crystal :", crystal.name)
    hc.output = output_path / "test037_test1.root"
    hc.attributes = [
        "PostPosition",
        "TotalEnergyDeposit",
        "TrackVolumeCopyNo",
        "PostStepUniqueVolumeID",
        "PreStepUniqueVolumeID",
        "GlobalTime",
        # "KineticEnergy", # not needed
        # "ProcessDefinedStep", # not needed
    ]

    # singles collection
    sc = sim.add_actor("DigitizerAdderActor", "Singles")
    sc.mother = crystal.name
    sc.input_digi_collection = "Hits"
    # sc.policy = "EnergyWinnerPosition"
    sc.policy = "EnergyWeightedCentroidPosition"
    # the following attributes is not needed in singles
    # sc.skip_attributes = ["KineticEnergy"]
    sc.output = hc.output

    # timing
    sim.run_timing_intervals = [[0, 0.0002 * sec]]



import opengate as gate
import opengate_core as g4
import re
from box import Box


class ElementBuilder:
    """
    A description of a G4Element that can be build.
    """

    def __init__(self, material_database):
        self.type = "element"
        self.name = None
        self.symbol = None
        self.Zeff = None
        self.Aeff = None
        self.material_database = material_database

    def __del__(self):
        pass

    def __repr__(self):
        u = gate.g4_units("g/mole")
        s = f"({self.type}) {self.name} ({self.symbol}) Z={self.Zeff} A={self.Aeff / u} g/mole"
        return s

    def read(self, line):
        self.type = "element"
        s = line.split(":")
        # name
        self.name = s[0]
        s = s[1].strip()
        s = s.split(";")
        # symbol
        self.symbol = gate.read_tag(s[0], "S")
        # Z
        self.Zeff = float(gate.read_tag(s[1], "Z"))
        # A with units
        self.Aeff = gate.read_tag_with_unit(s[2], "A")

    def build(self):
        m = g4.G4Element(self.name, self.symbol, self.Zeff, self.Aeff)
        # FIXME alternative with Build an element from isotopes via AddIsotope ?
        return m
